"""
An implementation of the profile connection scan algorithm.

Problem description:
Given
1. a static network for pedestrian routing
2. a temporal network with elementary transit events (accompanied with trip_ids)
3. source stop
4. target stop
5. interval start time
6. interval end time

Compute the pareto optimal departure times (from the source stop) and arrival times (to the target stop).
Considering pareto-optimality the following are beneficial
LATER (greater) departure time
EARLIER (smaller) arrival time

Now, the following departure_time, arrival_time pairs would all be pareto-optimal:
1, 3
3, 4
4, 5

However, e.g. (2, 4) would not be a pareto-optimal (departure_time, arrival_time) pair as it is dominated by (4,5)

while only one link in the static network can be traversed at a time.

Implements
"""
from collections import defaultdict

import networkx

from gtfspy.routing.connection import Connection
from gtfspy.routing.label import LabelTime
from gtfspy.routing.node_profile_simple import NodeProfileSimple
from gtfspy.routing.abstract_routing_algorithm import AbstractRoutingAlgorithm
from gtfspy.routing.pseudo_connections import compute_pseudo_connections
from gtfspy.routing.node_profile_c import NodeProfileC


class PseudoConnectionScanProfiler(AbstractRoutingAlgorithm):
    """
    Implementation of the profile connection scan algorithm presented in

    http://i11www.iti.uni-karlsruhe.de/extra/publications/dpsw-isftr-13.pdf
    """

    def __init__(self,
                 transit_events,
                 target_stop,
                 start_time=None,
                 end_time=None,
                 transfer_margin=0,
                 walk_network=None,
                 walk_speed=1.5,
                 verbose=False):
        """
        Parameters
        ----------
        transit_events: list[Connection]
            events are assumed to be ordered in DECREASING departure_time (!)
        target_stop: int
            index of the target stop
        start_time : int, optional
            start time in unixtime seconds
        end_time: int, optional
            end time in unixtime seconds (no connections will be scanned after this time)
        transfer_margin: int, optional
            required extra margin required for transfers in seconds
        walk_speed: float, optional
            walking speed between stops in meters / second.
        walk_network: networkx.Graph, optional
            each edge should have the walking distance as a data attribute ("distance_shape") expressed in meters
        verbose: boolean, optional
            whether to print out progress
        """
        AbstractRoutingAlgorithm.__init__(self)

        self._target = target_stop
        self._transit_connections = transit_events
        if start_time is None:
            start_time = transit_events[-1].departure_time
        if end_time is None:
            end_time = transit_events[0].departure_time
        self._start_time = start_time
        self._end_time = end_time
        self._transfer_margin = transfer_margin
        if walk_network is None:
            walk_network = networkx.Graph()
        self._walk_network = walk_network
        self._walk_speed = float(walk_speed)
        self._verbose = verbose

        # algorithm internals

        # trip flags:
        self.__trip_min_arrival_time = defaultdict(lambda: float("inf"))

        # initialize stop_profiles
        self._stop_profiles = defaultdict(lambda: NodeProfileC())
        # initialize stop_profiles for target stop, and its neighbors
        self._stop_profiles[self._target] = NodeProfileC(0)
        if target_stop in walk_network.nodes():
            for target_neighbor in walk_network.neighbors(target_stop):
                edge_data = walk_network.get_edge_data(target_neighbor, target_stop)
                walk_duration = edge_data["d_walk"] / self._walk_speed
                self._stop_profiles[target_neighbor] = NodeProfileC(walk_duration)
        pseudo_connection_set = compute_pseudo_connections(transit_events, self._start_time, self._end_time,
                                                           self._transfer_margin, self._walk_network,
                                                           self._walk_speed)
        self._pseudo_connections = list(pseudo_connection_set)
        self._all_connections = self._pseudo_connections + self._transit_connections
        self._all_connections.sort(key=lambda connection: -connection.departure_time)

    def _run(self):
        # if source node in s1:
        previous_departure_time = float("inf")
        connections = self._all_connections # list[Connection]
        n_connections_tot = len(connections)
        for i, connection in enumerate(connections):
            # basic checking + printing progress:
            if self._verbose and i % 1000 == 0:
                print(i, "/", n_connections_tot)
            assert (isinstance(connection, Connection))
            assert (connection.departure_time <= previous_departure_time)
            previous_departure_time = connection.departure_time

            # get all different "accessible" / arrival times (Pareto-optimal sets)
            arrival_profile = self._stop_profiles[connection.arrival_stop]  # NodeProfileSimple

            # Three possibilities:

            # 1. earliest arrival time (Profiles) via transfer
            earliest_arrival_time_via_transfer = arrival_profile.evaluate_earliest_arrival_time_at_target(
                connection.arrival_time, self._transfer_margin
            )

            # 2. earliest arrival time within same trip (equals float('inf') if not reachable)
            earliest_arrival_time_via_same_trip = self.__trip_min_arrival_time[connection.trip_id]

            # then, take the minimum (or the Pareto-optimal set) of these three alternatives.
            min_arrival_time = min(earliest_arrival_time_via_same_trip,
                                   earliest_arrival_time_via_transfer)

            # If there are no 'labels' to progress, nothing needs to be done.
            if min_arrival_time == float("inf"):
                continue

            # Update information for the trip
            if (not connection.is_walk) and (earliest_arrival_time_via_same_trip > min_arrival_time):
                self.__trip_min_arrival_time[connection.trip_id] = earliest_arrival_time_via_transfer

            # Compute the new "best" pareto_tuple possible (later: merge the sets of pareto-optimal labels)
            pareto_tuple = LabelTime(connection.departure_time, min_arrival_time)

            # update departure stop profile (later: with the sets of pareto-optimal labels)
            self._stop_profiles[connection.departure_stop].update_pareto_optimal_tuples(pareto_tuple)

    @property
    def stop_profiles(self):
        """
        Returns
        -------
        _stop_profiles : dict[int, NodeProfileSimple]
            The pareto tuples necessary.
        """
        assert self._has_run
        return self._stop_profiles



import unittest

from gtfspy.routing.forwardjourney import ForwardJourney
from gtfspy.routing.connection import Connection


class ForwardJourneyTest(unittest.TestCase):

    def test_add_leg(self):
        journey = ForwardJourney()
        leg1 = Connection(departure_stop=0, arrival_stop=1, departure_time=0, arrival_time=1,
                          trip_id="tripI", seq=1, is_walk=False)
        journey.add_leg(leg1)
        self.assertEqual(len(journey.legs), 1)
        self.assertEqual(journey.departure_time, leg1.departure_time)
        self.assertEqual(journey.arrival_time, leg1.arrival_time)
        self.assertEqual(journey.n_boardings, 1)

        leg2 = Connection(departure_stop=1, arrival_stop=2, departure_time=1, arrival_time=2,
                          trip_id="tripI", seq=1, is_walk=False)
        journey.add_leg(leg2)
        self.assertEqual(len(journey.legs), 2)
        self.assertEqual(journey.departure_time, leg1.departure_time)
        self.assertEqual(journey.arrival_time, leg2.arrival_time)
        self.assertEqual(journey.n_boardings, 1)

    def test_dominates(self):
        leg1 = Connection(departure_stop=0, arrival_stop=1, departure_time=0, arrival_time=1,
                          trip_id="tripI", seq=1, is_walk=False)
        leg2 = Connection(departure_stop=1, arrival_stop=2, departure_time=1, arrival_time=2,
                          trip_id="tripI", seq=1, is_walk=False)
        leg3 = Connection(departure_stop=1, arrival_stop=2, departure_time=1, arrival_time=3,
                          trip_id="tripI", seq=1, is_walk=False)
        journey1 = ForwardJourney(legs=[leg1])
        journey2 = ForwardJourney(legs=[leg2])
        journey12 = ForwardJourney(legs=[leg1, leg2])
        journey13 = ForwardJourney(legs=[leg1, leg3])
        self.assertTrue(journey12.dominates(journey13))
        self.assertFalse(journey1.dominates(journey2))
        self.assertTrue(journey1.dominates(journey1, consider_time=False, consider_boardings=False))

    def test_basics(self):
        event_list_raw_data = [
            (0, 1, 0, 10, "trip_1", 1),
            (1, 100, 32, 36, "trip_5", 1),
            (100, 3, 36, 40, "trip_5", 2),
            (3, 4, 40, 41, "trip_4", 1),
            (4, 2, 44, 50, None, 1)
        ]
        legs = list(map(lambda el: Connection(*el), event_list_raw_data))
        test_journey = ForwardJourney(legs)

        self.assertIsInstance(test_journey, ForwardJourney)
        self.assertIsInstance(test_journey.get_legs(), list)

        self.assertEqual(test_journey.n_boardings, 3)
        self.assertEqual(test_journey.get_transfers(), 2)
        self.assertEqual(test_journey.get_travel_time(), 50)
        self.assertIsInstance(test_journey.get_waiting_times(), list)
        self.assertEqual(test_journey.get_total_waiting_time(), 22 + 0 + 3)
        self.assertEqual(len(test_journey.get_all_stops()), 6)

    def test_transfer_stop_pairs(self):
        event_list_raw_data = [
            (0, 1, 0, 10, "trip_1", 1),
            (1, 100, 32, 36, "trip_5", 1),
            (100, 3, 36, 40, "trip_5", 2),
            (3, 4, 40, 41, "trip_4", 1),
            (4, 2, 44, 50, None, 1),
            (10, 11, 52, 55, "trip_6", 1)
        ]
        legs = list(map(lambda el: Connection(*el), event_list_raw_data))
        test_journey = ForwardJourney(legs)

        transfer_stop_pairs = test_journey.get_transfer_stop_pairs()
        print(transfer_stop_pairs)
        self.assertEqual(len(transfer_stop_pairs), 3)
        self.assertEqual(transfer_stop_pairs[0][0], 1)
        self.assertEqual(transfer_stop_pairs[0][1], 1)
        self.assertEqual(transfer_stop_pairs[1][0], 3)
        self.assertEqual(transfer_stop_pairs[1][1], 3)
        self.assertEqual(transfer_stop_pairs[2][0], 2)
        self.assertEqual(transfer_stop_pairs[2][1], 10)




from __future__ import print_function

import os
import sqlite3
import unittest

from gtfspy.gtfs import GTFS
from gtfspy.import_gtfs import import_gtfs


# noinspection PyTypeChecker
class TestImport(unittest.TestCase):

    @classmethod
    def setup_class(cls):
        """This method is run once for each class before any tests are run"""
        pass

    @classmethod
    def teardown_class(cls):
        """This method is run once for each class _after_ all tests are run"""
        pass

    def tearDown(self):
        # disconnect from the existing in memory connection
        self.conn.close()

    def setUp(self):
        """This method is run once before _each_ test method is executed"""
        self.conn = sqlite3.connect(':memory:')
        self.agencyText = \
            'agency_id, agency_name, agency_timezone, agency_url' \
            '\n ag1, CompNet, Europe/Zurich, www.example.com'
        self.stopsText = \
            'stop_id, stop_name, stop_lat, stop_lon, parent_station' \
            '\nSID1, "Parent-Stop-Name", 1.0, 2.0, ' \
            '\nSID2, Boring Stop Name, 1.1, 2.1, SID1' \
            '\n3, Boring Stop Name1, 1.2, 2.2, ' \
            '\n4, Boring Stop Name2, 1.3, 2.3, 3' \
            '\n5, StopCloseToFancyStop, 1.0001, 2.0001, ' \
            '\nT1, t1, 1.0001, 2.2, ' \
            '\nT2, t2, 1.0002, 2.2, ' \
            '\nT3, t3, 1.00015, 2.2, ' \
            '\nT4, t4, 1.0001, 2.2, '
        self.calendarText = \
            'service_id, monday, tuesday, wednesday, thursday, friday, saturday, sunday,' \
            'start_date, end_date' \
            '\nservice1, 1, 1, 1, 1, 1, 1, 1, 20160321, 20160327' \
            '\nservice2, 0, 0, 0, 0, 0, 0, 0, 20160321, 20160327' \
            '\nfreq_service, 1, 1, 1, 1, 1, 1, 1, 20160329, 20160329'
        self.calendarDatesText = \
            'service_id, date, exception_type' \
            '\nservice1, 20160322, 2' \
            '\nextra_service, 20160321, 1' \
            '\nservice2, 20160322, 1' \
            '\nphantom_service, 20160320, 2'
        # 1 -> service added
        # 2 -> service removed
        # note some same service IDs as in self.calendarText
        self.tripText = \
            "route_id, service_id, trip_id, trip_headsign, trip_short_name, shape_id" \
            "\nservice1_route, service1, service1_trip1, going north, trip_s1t1, shape_s1t1" \
            "\nservice2_route, service2, service2_trip1, going north, trip_s2t1, shape_s2t1" \
            "\nes_route, extra_service, es_trip1, going north, trip_es1, shape_es1" \
            "\nfrequency_route, freq_service, freq_trip_scheduled, going north, freq_name, shape_es1"
        self.routesText = \
            "route_id, agency_id, route_short_name, route_long_name, route_type" \
            "\nservice1_route, ag1, r1, route1, 0" \
            "\nservice2_route, ag1, r2, route2, 1" \
            "\nfrequency_route, ag1, freq_route, schedule frequency route, 2"
        self.shapeText = \
            "shape_id, shape_pt_lat, shape_pt_lon, shape_pt_sequence" \
            "\n shape_s1t1,1.0,2.0,0" \
            "\n shape_s1t1,1.001,2.0,1" \
            "\n shape_s1t1,1.001,2.001,10" \
            "\n shape_s1t1,1.10001,2.10001,100"
        self.stopTimesText = \
            "trip_id, arrival_time, departure_time, stop_sequence, stop_id" \
            "\nservice1_trip1,0:06:10,0:06:10,0,SID1" \
            "\nservice1_trip1,0:06:15,0:06:16,1,SID2" \
            "\nfreq_trip_scheduled,0:00:00,0:00:00,1,SID1" \
            "\nfreq_trip_scheduled,0:02:00,0:02:00,1,SID2"
        self.frequenciesText = \
            "trip_id, start_time, end_time, headway_secs, exact_times" \
            "\nfreq_trip_scheduled, 14:00:00, 16:00:00, 600, 1"
        self.transfersText = \
            "from_stop_id, to_stop_id, transfer_type, min_transfer_time" \
            "\nT1, T2, 0, " \
            "\nT2, T3, 1, " \
            "\nT3, T1, 2, 120" \
            "\nT1, T4, 3, "
        self.feedInfoText = \
            "feed_publisher_name, feed_publisher_url, feed_lang, feed_start_date, feed_end_date, feed_version" \
            "\nThePublisher, www.example.com, en, 20160321, 20160327, 1.0"

        self.fdict = {
            'agency.txt':           self.agencyText,
            'stops.txt':            self.stopsText,
            'calendar.txt':         self.calendarText,
            'calendar_dates.txt':   self.calendarDatesText,
            'trips.txt':            self.tripText,
            'routes.txt':           self.routesText,
            'shapes.txt':           self.shapeText,
            'stop_times.txt':       self.stopTimesText,
            'frequencies.txt':      self.frequenciesText,
            'transfers.txt':        self.transfersText,
            'feed_info.txt':        self.feedInfoText
        }
        self.orig_row_factory = self.conn.row_factory

    def setDictConn(self):
        prev_factory = self.conn.row_factory

        def dict_factory(cursor, row):
            d = {}
            for idx, col in enumerate(cursor.description):
                d[col[0]] = row[idx]
            return d

        self.conn.row_factory = dict_factory
        return prev_factory

    def setRowConn(self):
        prev_factory = self.conn.row_factory
        self.conn.row_factory = self.orig_row_factory
        return prev_factory

    def printTable(self, table_name):
        """
        Pretty prints a table with name table_name.

        Parameters
        ----------
        table_name : str
            name of the table
        """
        prev_row_factory = self.setRowConn()
        print("")
        print("table " + table_name)
        print("-------------------")
        cur = self.conn.execute("SELECT * FROM %s" % table_name)
        names = [d[0] for d in cur.description]
        for name in names:
            print(name + ', ', end="")
        print("")
        for row in cur:
            print(row)
        self.conn.row_factory = prev_row_factory

    def tearDown(self):
        """This method is run once after _each_ test method is executed"""
        pass

    # def test_download_date_importing(self):
    #     path = tempfile.mkdtemp(prefix='2015-02-03', dir="/tmp/2016-02-03/")
    #     print path
    #     tempfile.mkstemp(suffix='.sqlite', prefix='tmp', dir="/tmp/2015-02-03/")

    def test_stopLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        self.setDictConn()
        # sqlite returns now list of dicts
        rows = self.conn.execute("SELECT * FROM stops").fetchall()
        assert len(rows) > 4  # some data should be imported
        assert rows[0]['stop_I'] == 1
        # Store quotes in names:
        parent_index = None
        for i, row in enumerate(rows):
            if row['name'] == '"Parent-Stop-Name"':
                parent_index = i
                break
        assert parent_index is not None
        parent_stop_I = rows[parent_index]['stop_I']
        boring_index = None
        for i, row in enumerate(rows):
            if row['name'] == "Boring Stop Name":
                boring_index = i
                break
        assert boring_index is not None
        assert rows[boring_index]['parent_I'] == parent_stop_I
        assert rows[boring_index]['self_or_parent_I'] == parent_stop_I
        assert rows[3]['self_or_parent_I'] == 3

    def test_agencyLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        self.conn.commit()
        cursor = self.conn.cursor()
        rows = cursor.execute("SELECT agency_id FROM agencies").fetchall()
        assert len(rows) == 1
        assert rows[0][0] == u'ag1', rows[0][0]

    def test_agencyLoaderTwoTimeZonesFail(self):
        newagencytext = \
            self.agencyText + "\n123, AgencyFromDifferentTZ, Europe/Helsinki, www.buahaha.com"
        self.fdict['agency.txt'] = newagencytext
        with self.assertRaises(ValueError):
            import_gtfs(self.fdict, self.conn, preserve_connection=True)

    def test_routeLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        assert len(self.conn.execute("SELECT * FROM routes").fetchall()) > 0

    def test_calendarLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        self.setDictConn()
        rows = self.conn.execute("SELECT * FROM calendar").fetchall()
        assert len(rows[0]) == 11
        for key in 'm t w th f s su start_date end_date service_id service_I'.split():
            assert key in rows[0], 'no key ' + key

    def test_calendarDatesLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        rows = self.conn.execute("SELECT * FROM calendar").fetchall()
        self.setDictConn()
        rows = self.conn.execute("SELECT * FROM calendar_dates").fetchall()
        for row in rows:
            assert isinstance(row['service_I'], int)
        # calendar table should be increased by two dummy row
        rows = self.conn.execute("SELECT * "
                                 "FROM calendar "
                                 "WHERE service_id='phantom_service'").fetchall()
        # Whether this should be the case is negotiable, though
        self.assertEqual(len(rows), 1, "phantom service should be present in the calendar")

    def test_tripLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        rows = self.conn.execute("SELECT * FROM trips").fetchall()
        self.assertGreaterEqual(len(rows), 1)

    def test_dayLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        # Now, there should be
        # a regular trip according to calendar dates without any exceptions:
        self.setDictConn()
        query1 = "SELECT trip_I " \
                 "FROM days " \
                 "JOIN trips " \
                 "USING(trip_I) " \
                 "JOIN calendar " \
                 "USING(service_I) " \
                 "WHERE date='2016-03-21'" \
                 "AND service_id='service1'"
        res = self.conn.execute(query1).fetchall()
        assert len(res) == 1
        trip_I_service_1 = res[0]['trip_I']
        print(trip_I_service_1)
        query2 = "SELECT * FROM days WHERE trip_I=%s" % trip_I_service_1
        self.assertEqual(len(self.conn.execute(query2).fetchall()), 6,
                         "There should be 6 days with the trip_I "
                         "corresponding to service_id service1")
        query3 = "SELECT * " \
                 "FROM days " \
                 "JOIN trips " \
                 "USING(trip_I) " \
                 "JOIN calendar " \
                 "USING(service_I) " \
                 "WHERE date='2016-03-22'" \
                 "AND service_id='service1'"
        self.assertEqual(len(self.conn.execute(query3).fetchall()), 0,
                         "There should be no trip on date 2016-03-22"
                         "for service1 due to calendar_dates")
        query4 = "SELECT date " \
                 "FROM days " \
                 "JOIN trips " \
                 "USING(trip_I) " \
                 "JOIN calendar " \
                 "USING(service_I) " \
                 "WHERE service_id='service2'"
        self.assertEqual(len(self.conn.execute(query4).fetchall()), 1,
                         "There should be only one trip for service 2")
        self.assertEqual(self.conn.execute(query4).fetchone()['date'], "2016-03-22",
                         "and the date should be 2016-03-22")
        query6 = "SELECT * " \
                 "FROM days " \
                 "JOIN trips " \
                 "USING(trip_I) " \
                 "JOIN calendar " \
                 "USING(service_I) " \
                 "WHERE service_id='phantom_service'"
        res = self.conn.execute(query6).fetchall()

        self.assertEqual(len(res), 0, "there should be no phantom trips due to phantom service"
                                      "even though phantom service is in calendar"
                         )

    def test_shapeLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        self.setDictConn()
        keys = ['shape_id', 'lat', 'lon', 'seq', 'd']
        table = self.conn.execute("SELECT * FROM shapes").fetchall()
        assert table[1]['d'] > 0, "distance traveled should be > 0"
        for key in keys:
            assert key in table[0], "key " + key + " not in shapes table"

    def test_stopTimesLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        self.setDictConn()
        stoptimes = self.conn.execute("SELECT * FROM stop_times").fetchall()
        keys = ['stop_I', 'shape_break', 'trip_I', 'arr_time',
                'dep_time', 'seq', 'arr_time_ds', 'dep_time_ds']
        for key in keys:
            assert key in stoptimes[0]
        assert stoptimes[0]['dep_time_ds'] == 370
        assert stoptimes[0]['shape_break'] == 0
        assert stoptimes[1]['shape_break'] == 3

    def test_stopDistancesLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        query = "SELECT * FROM stop_distances"
        # set dictionary like row connection:
        self.setDictConn()
        rows = self.conn.execute(query).fetchall()
        assert len(rows) > 0
        for row in rows:
            print(row)
            assert row['d'] >= 0, "distance should be defined for all pairs in the stop_distances table"

    def test_metaDataLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        try:
            query = "SELECT * FROM metadata"
            self.conn.execute(query)
        except AssertionError:
            assert False, "The database should have a table named metadata"

    def test_frequencyLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        # "\nfrequency_route, freq_service, freq_trip, going north, freq_name, shape_es1" \
        keys = ["trip_I", "start_time", "end_time", "headway_secs", "exact_times", "start_time_ds", "end_time_ds"]
        self.setDictConn()
        rows = self.conn.execute("SELECT * FROM frequencies").fetchall()
        for key in keys:
            row = rows[0]
            assert key in row
        for row in rows:
            if row["start_time_ds"] == 14 * 3600:
                self.assertEqual(row["exact_times"], 1)
        # there should be twelve trips with service_I freq
        count = self.conn.execute("SELECT count(*) AS count FROM trips JOIN calendar "
                                  "USING(service_I) WHERE service_id='freq_service'").fetchone()['count']

        assert count == 12, count
        rows = self.conn.execute("SELECT trip_I FROM trips JOIN calendar "
                                 "USING(service_I) WHERE service_id='freq_service'").fetchall()
        for row in rows:
            trip_I = row['trip_I']
            res = self.conn.execute("SELECT * FROM stop_times WHERE trip_I={trip_I}".format(trip_I=trip_I)).fetchall()
            assert len(res) > 1, res
        self.setRowConn()
        g = GTFS(self.conn)
        print("Stop times: \n\n ", g.get_table("stop_times"))
        print("Frequencies: \n\n ", g.get_table("frequencies"))
        # should there be more tests?
        # check that the original trip_id does not exist in frequencies, trips, or stop_times?

    def test_transfersLoader(self):
        """
        First tests that the basic import to the transfers table is correct, and then checks that
        the information from transfers.txt is also flows to the stop_distances table.
        """
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        keys = ["from_stop_I", "to_stop_I", "transfer_type", "min_transfer_time"]
        self.setDictConn()
        transfers = self.conn.execute("SELECT * FROM transfers").fetchall()
        for key in keys:
            transfer = transfers[0]
            assert key in transfer

        from_stop_I_no_transfer = None
        to_stop_I_no_transfer = None
        from_stop_I_timed_transfer = None
        to_stop_I_timed_transfer = None
        from_stop_I_min_transfer = None
        to_stop_I_min_transfer = None
        min_transfer_time_min_transfer = None

        for transfer in transfers:
            transfer_type = transfer["transfer_type"]
            from_stop_I = transfer['from_stop_I']
            to_stop_I = transfer['to_stop_I']
            min_transfer_time = transfer["min_transfer_time"]
            assert isinstance(from_stop_I, int)
            assert isinstance(to_stop_I, int)
            assert isinstance(transfer_type, int)
            assert isinstance(min_transfer_time, int) or (min_transfer_time is None)
            if transfer["transfer_type"] == 3:  # no transfer
                from_stop_I_no_transfer = from_stop_I
                to_stop_I_no_transfer = to_stop_I
            elif transfer["transfer_type"] == 2:
                from_stop_I_min_transfer = from_stop_I
                to_stop_I_min_transfer = to_stop_I
                min_transfer_time_min_transfer = min_transfer_time
            elif transfer["transfer_type"] == 1:
                from_stop_I_timed_transfer = from_stop_I
                to_stop_I_timed_transfer = to_stop_I

        base_query = "SELECT * FROM stop_distances WHERE from_stop_I=? and to_stop_I=?"
        # no_transfer
        no_transfer_rows = self.conn.execute(base_query, (from_stop_I_no_transfer, to_stop_I_no_transfer)).fetchall()
        assert len(no_transfer_rows) == 0
        timed_transfer_rows = \
            self.conn.execute(base_query, (from_stop_I_timed_transfer, to_stop_I_timed_transfer)).fetchall()
        assert len(timed_transfer_rows) == 1
        assert timed_transfer_rows[0]['min_transfer_time'] == 0
        min_transfer_rows = \
            self.conn.execute(base_query, (from_stop_I_min_transfer, to_stop_I_min_transfer)).fetchall()
        assert len(min_transfer_rows) == 1
        assert min_transfer_rows[0]['min_transfer_time'] == min_transfer_time_min_transfer

    def test_feedInfoLoader(self):
        import_gtfs(self.fdict, self.conn, preserve_connection=True)
        keys = ["feed_publisher_name", "feed_publisher_url", "feed_lang", "feed_start_date", "feed_end_date"]
        self.setDictConn()
        rows = self.conn.execute("SELECT * FROM feed_info").fetchall()
        for key in keys:
            row = rows[0]
            assert key in row

    def test_testDataImport(self):
        gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        import_gtfs(gtfs_source_dir, self.conn, preserve_connection=True)

    def test_zipImport(self):
        gtfs_source_zip = os.path.join(os.path.dirname(__file__), "test_data/test_gtfs.zip")
        import_gtfs(gtfs_source_zip, self.conn, preserve_connection=True)

    def test_importMultiple(self):
        gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")

        gtfs_sources = [gtfs_source_dir, self.fdict]
        error_raised = False
        try:
            import_gtfs(gtfs_sources, self.conn, preserve_connection=True)
        except ValueError:
            error_raised = True
        assert error_raised, "different timezones in multiple feeds should raise an error"


        #mod_agencyText = \
        #    'agency_id, agency_name, agency_timezone, agency_url' \
         #   '\nag1, CompNet, America/Los_Angeles, www.example.com'
        #self.fdict['agency.txt'] = mod_agencyText


        # test that if trip_id:s (or stop_id:s etc. ) are the same in two feeds,
        # they get different trip_Is in the database created

        self.tearDown()
        self.setUp()

        # assert if importing two of the same feed will create the double number of trips
        gtfs_source = [self.fdict]
        import_gtfs(gtfs_source, self.conn, preserve_connection=True)
        n_rows_ref = self.conn.execute("SELECT count(*) FROM trips").fetchone()[0]
        self.tearDown()
        self.setUp()
        gtfs_sources = [self.fdict, self.fdict]
        import_gtfs(gtfs_sources, self.conn, preserve_connection=True)
        n_rows_double = self.conn.execute("SELECT count(*) FROM trips").fetchone()[0]
        self.assertEqual(n_rows_double, 2*n_rows_ref)

        # check for duplicate trip_I's
        rows = self.conn.execute("SELECT count(*) FROM trips GROUP BY trip_I").fetchall()
        for row in rows:
            self.assertIs(row[0],1)

        # check for duplicate service_I's in calendar
        rows = self.conn.execute("SELECT count(*) FROM calendar GROUP BY service_I").fetchall()
        for row in rows:
            self.assertIs(row[0], 1)

        # check for duplicate service_I's in calendar_dates
        rows = self.conn.execute("SELECT count(*) FROM calendar_dates GROUP BY service_I").fetchall()
        for row in rows:
            self.assertIs(row[0], 1)

        # check for duplicate route_I's
        rows = self.conn.execute("SELECT count(*) FROM routes GROUP BY route_I").fetchall()
        for row in rows:
            self.assertIs(row[0], 1)

        # check for duplicate agency_I's
        rows = self.conn.execute("SELECT count(*) FROM agencies GROUP BY agency_I").fetchall()
        for row in rows:
            self.assertIs(row[0], 1)

        # check for duplicate stop_I's
        rows = self.conn.execute("SELECT count(*) FROM stops GROUP BY stop_I").fetchall()
        for row in rows:
            self.assertIs(row[0], 1)

    def test_sources_required(self):
        self.fdict.pop("stops.txt")
        with self.assertRaises(AssertionError):
            import_gtfs(self.fdict, self.conn)

    def test_sources_required_multiple(self):
        fdict_copy = dict(self.fdict)
        fdict_copy.pop("stops.txt")
        with self.assertRaises(AssertionError):
            import_gtfs([self.fdict, fdict_copy], self.conn)

    def test_resequencing_stop_times(self):
        gtfs_source = self.fdict.copy()
        gtfs_source.pop('stop_times.txt')

        gtfs_source['stop_times.txt'] = \
            self.stopTimesText = \
            "trip_id, arrival_time, departure_time, stop_sequence, stop_id" \
            "\nservice1_trip1,0:06:10,0:06:10,0,SID1" \
            "\nservice1_trip1,0:06:15,0:06:16,10,SID2" \
            "\nfreq_trip_scheduled,0:00:00,0:00:00,1,SID1" \
            "\nfreq_trip_scheduled,0:02:00,0:02:00,123,SID2"
        import_gtfs(gtfs_source, self.conn, preserve_connection=True)

        rows = self.conn.execute("SELECT seq FROM stop_times ORDER BY trip_I, seq").fetchall()
        for row in rows:
            print(row)
        self.assertEqual(rows[0][0], 1)
        self.assertEqual(rows[1][0], 2)
        self.assertEqual(rows[2][0], 1)
        self.assertEqual(rows[3][0], 2)

    def test_metaData(self):
        # TODO! untested
        pass



from __future__ import unicode_literals

import datetime
import os
import sqlite3
import unittest

import numpy
from six import string_types

import pandas

from gtfspy.gtfs import GTFS
from gtfspy.util import wgs84_distance
from gtfspy.route_types import BUS, TRAM, ALL_ROUTE_TYPES


class TestGTFS(unittest.TestCase):
    """
    Test data can be found under test_data/ directory.
    """

    @classmethod
    def setUpClass(cls):
        """ This method is run once before executing any tests"""
        cls.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        cls.G = GTFS.from_directory_as_inmemory_db(cls.gtfs_source_dir)

    def setUp(self):
        """This method is run once before _each_ test method is executed"""
        self.gtfs_source_dir = self.__class__.gtfs_source_dir
        self.gtfs = self.__class__.G

    def tearDown(self):
        """This method is run once after _each_ test method is executed"""
        pass

    def test_get_cursor(self):
        self.assertTrue(isinstance(self.gtfs.get_cursor(), sqlite3.Cursor))

    def test_get_timezone_name(self):
        self.assertIsInstance(self.gtfs.get_timezone_string(), str)

    def test_get_day_start_ut(self):
        """ America/Los_Angeles on 2016,1,1 should map to -07:00 when DST IS in place"""
        date = datetime.datetime(2016, 6, 15)
        epoch = datetime.datetime(1970, 1, 1)
        day_start_utc_ut = (date - epoch).total_seconds()
        # day starts 7 hours later in Los Angeles
        day_start_ut_should_be = day_start_utc_ut + 7 * 3600
        day_start_ut_is = self.gtfs.get_day_start_ut(date)
        self.assertEquals(day_start_ut_should_be, day_start_ut_is)

    def test_get_main_database_path(self):
        self.assertEqual(self.gtfs.get_main_database_path(),  "", "path of an in-memory database should equal ''")

        from gtfspy.import_gtfs import import_gtfs
        try:
            fname = self.gtfs_source_dir + "/test_gtfs.sqlite"
            if os.path.exists(fname) and os.path.isfile(fname):
                os.remove(fname)
            conn = sqlite3.connect(fname)
            import_gtfs(self.gtfs_source_dir, conn, preserve_connection=True, print_progress=False)
            G = GTFS(conn)
            self.assertTrue(os.path.exists(G.get_main_database_path()))
            self.assertIn(u"/test_gtfs.sqlite", G.get_main_database_path(), "path should be correct")
        finally:
            if os.path.exists(fname) and os.path.isfile(fname):
                os.remove(fname)

    def test_get_table(self):
        df = self.gtfs.get_table(u"agencies")
        self.assertTrue(isinstance(df, pandas.DataFrame))

    def test_get_table_names(self):
        tables = self.gtfs.get_table_names()
        self.assertTrue(isinstance(tables, list))
        self.assertGreater(len(tables), 11, u"quite many tables should be available")
        self.assertIn(u"routes", tables)

    def test_get_all_route_shapes(self):
        res = self.gtfs.get_all_route_shapes()
        self.assertTrue(isinstance(res, list))
        el = res[0]
        keys = u"name type agency lats lons".split()
        for key in keys:
            self.assertTrue(key in el)

        for el in res:
            self.assertTrue(isinstance(el[u"name"], string_types), type(el[u"name"]))
            self.assertTrue(isinstance(el[u"type"], (int, numpy.int_)), type(el[u'type']))
            self.assertTrue(isinstance(el[u"agency"], string_types))
            self.assertTrue(isinstance(el[u"lats"], list), type(el[u'lats']))
            self.assertTrue(isinstance(el[u"lons"], list))
            self.assertTrue(isinstance(el[u'lats'][0], float))
            self.assertTrue(isinstance(el[u'lons'][0], float))

    def test_get_shape_distance_between_stops(self):
        # tested as a part of test_to_directed_graph, although this could be made a separate test as well
        pass

    def test_stops(self):
        self.assertIsInstance(self.gtfs.stops(), pandas.DataFrame)

    def test_stops_by_route_type(self):
        stops = self.gtfs.get_stops_for_route_type(BUS)
        self.assertEquals(len(stops), len(self.gtfs.stops()))
        self.assertIsInstance(stops, pandas.DataFrame)
        stops = self.gtfs.get_stops_for_route_type(TRAM)
        self.assertEquals(len(stops), 0)

    def test_get_timezone_string(self):
        tz_string = self.gtfs.get_timezone_string()
        self.assertEquals(len(tz_string), 5)
        self.assertIn(tz_string[0], "+-")
        for i in range(1, 5):
            self.assertIn(tz_string[i], "0123456789")
        dt = datetime.datetime(1970, 1, 1)
        tz_string_epoch = self.gtfs.get_timezone_string(dt)
        # self.assertEqual(tz_string, tz_string_epoch)

    def test_timezone_conversions(self):
        """
        Two methods are tested:
            ut_seconds_to_gtfs_datetime
            unlocalized_datetime_to_ut_seconds
        """
        ut = 10.0
        gtfs_dt = self.gtfs.unixtime_seconds_to_gtfs_datetime(ut)
        unloc_dt = gtfs_dt.replace(tzinfo=None)
        ut_later = self.gtfs.unlocalized_datetime_to_ut_seconds(unloc_dt)
        self.assertTrue(ut == ut_later)

    def test_get_trip_trajectory_data_within_timespan(self):
        # untested, really
        s, e = self.gtfs.get_approximate_schedule_time_span_in_ut()
        res = self.gtfs.get_trip_trajectories_within_timespan(s, s + 3600 * 24)
        self.assertTrue(isinstance(res, dict))
        # TODO! Not properly tested yet.

    def test_get_stop_count_data(self):
        dt_start_query = datetime.datetime(2007, 1, 1, 7, 59, 59)
        dt_end_query = datetime.datetime(2007, 1, 1, 10, 2, 1)
        start_query = self.gtfs.unlocalized_datetime_to_ut_seconds(dt_start_query)
        end_query = self.gtfs.unlocalized_datetime_to_ut_seconds(dt_end_query)
        df = self.gtfs.get_stop_count_data(start_query, end_query)
        self.assertTrue(isinstance(df, pandas.DataFrame))
        columns = ["stop_I", "count", "lat", "lon", "name"]
        for c in columns:
            self.assertTrue(c in df.columns)
            el = df[c].iloc[0]
            if c in ["stop_I", "count"]:
                self.assertTrue(isinstance(el, (int, numpy.int_)))
            if c in ["lat", "lon"]:
                self.assertTrue(isinstance(el, float))
            if c in ["name"]:
                self.assertTrue(isinstance(el, string_types), type(el))
        self.assertTrue((df['count'].values > 0).any())

    def test_get_segment_count_data(self):
        dt_start_query = datetime.datetime(2007, 1, 1, 7, 59, 59)
        dt_end_query = datetime.datetime(2007, 1, 1, 10, 2, 1)
        start_query = self.gtfs.unlocalized_datetime_to_ut_seconds(dt_start_query)
        end_query = self.gtfs.unlocalized_datetime_to_ut_seconds(dt_end_query)

        res = self.gtfs.get_segment_count_data(start_query, end_query, use_shapes=True)
        self.assertGreater(len(res), 0)
        self.assertIsNotNone(res, "this is a 'it compiles' test")

    def test_get_tripIs_active_in_range(self):
        dt_start_query = datetime.datetime(2007, 1, 1, 7, 59, 59)
        dt_end_query = datetime.datetime(2007, 1, 1, 8, 2, 1)
        dt_start_real = datetime.datetime(2007, 1, 1, 8, 0, 00)
        dt_end_real = datetime.datetime(2007, 1, 1, 8, 10, 00)
        start_query = self.gtfs.unlocalized_datetime_to_ut_seconds(dt_start_query)
        end_query = self.gtfs.unlocalized_datetime_to_ut_seconds(dt_end_query)
        start_real = self.gtfs.unlocalized_datetime_to_ut_seconds(dt_start_real)
        end_real = self.gtfs.unlocalized_datetime_to_ut_seconds(dt_end_real)

        found = False
        df = self.gtfs.get_tripIs_active_in_range(start_query, end_query)
        self.assertGreater(len(df), 0)
        for row in df.itertuples():
            self.assertTrue((row.start_time_ut <= end_query) and \
                   (row.end_time_ut >= start_query), "some trip does not overlap!")
            if row.start_time_ut == start_real and row.end_time_ut == end_real:
                found = True
                # check that overlaps
        # for debugging::
        # if not found:
        #     print start_real, end_real
        #     for h in header:
        #         print h + " |",
        #     print ""
        #     for row in rows:
        #         for el in row:
        #             print el,
        #         print ""
        self.assertTrue(found, "a trip that should be found is not found")

    def test_get_trip_counts_per_day(self):
        df = self.gtfs.get_trip_counts_per_day()
        columns = "date_str trip_counts".split(" ")
        for c in columns:
            self.assertTrue(c in df.columns)
        el = df.iloc[0]
        self.assertIsInstance(el["date_str"], string_types)
        self.assertIsInstance(el["trip_counts"], (int, numpy.int_))

    def test_get_spreading_trips(self):
        pass  # untested

    def test_get_closest_stop(self):
        # First row in test_data:
        # FUR_CREEK_RES, Furnace Creek Resort (Demo),, 36.425288, -117.133162,,
        lat_s, lon_s = 36.425288, -117.133162
        lat, lon = lat_s + 10**-5, lon_s + 10**-5
        stop_I = self.gtfs.get_closest_stop(lat, lon)
        self.assertTrue(isinstance(stop_I, int))
        df = self.gtfs.stop(stop_I)
        name = df['name'][0]
        # print name
        # check that correct stop has been found:
        self.assertTrue(name == "Furnace Creek Resort (Demo)")
        # distance to the stop should be below 50 meters for such a small separation:
        self.assertTrue(wgs84_distance(lat, lon, lat_s, lon_s) < 50)

    def test_get_route_name_and_type_of_tripI(self):
        # just a simple random test:
        trip_I = 1
        name, type_ = self.gtfs.get_route_name_and_type_of_tripI(trip_I)
        self.assertTrue(isinstance(name, string_types))
        self.assertTrue(isinstance(type_, int))

    def test_get_trip_stop_time_data(self):
        start_ut, end_ut = self.gtfs.get_approximate_schedule_time_span_in_ut()
        dsut_dict = self.gtfs.get_tripIs_within_range_by_dsut(start_ut, end_ut)
        dsut, trip_Is = list(dsut_dict.items())[0]
        df = self.gtfs.get_trip_stop_time_data(trip_Is[0], dsut)
        self.assertTrue(isinstance(df, pandas.DataFrame))
        columns = u"dep_time_ut lat lon seq shape_break".split(" ")
        el = df.iloc[0]
        for c in columns:
            self.assertTrue(c in df.columns)
            if c in u"dep_time_ut lat lon".split(" "):
                self.assertTrue(isinstance(el[c], float))
            if c in u"seq".split(" "):
                self.assertTrue(isinstance(el[c], (int, numpy.int_)), type(el[c]))

    def test_get_straight_line_transfer_distances(self):
        data = self.gtfs.get_straight_line_transfer_distances()
        a_stop_I = None
        for index, row in data.iterrows():
            self.assertTrue(row[u'from_stop_I'] is not None)
            a_stop_I = row[u'from_stop_I']
            self.assertTrue(row[u'to_stop_I'] is not None)
            self.assertTrue(row[u'd'] is not None)
        data = self.gtfs.get_straight_line_transfer_distances(a_stop_I)
        self.assertGreater(len(data), 0)

    def test_get_conservative_gtfs_time_span_in_ut(self):
        start_ut, end_ut = self.gtfs.get_approximate_schedule_time_span_in_ut()
        start_dt = datetime.datetime(2007, 1, 1)
        start_ut_comp = self.gtfs.unlocalized_datetime_to_ut_seconds(start_dt)
        end_dt = datetime.datetime(2010, 12, 31)
        end_ut_comp = self.gtfs.unlocalized_datetime_to_ut_seconds(end_dt) + (28 * 3600)
        self.assertTrue(start_ut == start_ut_comp)
        self.assertTrue(end_ut == end_ut_comp)

    def test_get_location_name(self):
        location_name = self.G.get_location_name()
        # TODO: is this location name retrieval working properly now?
        # self.assertEqual(location_name, "test_data")
        self.assertTrue(isinstance(location_name, string_types), type(location_name))
        self.assertGreater(len(location_name), 0)

    def test_get_n_stop_counts(self):
        self.assertEqual(self.G.get_n_stops(), 9)

    def test_get_modes(self):
        modes = self.G.get_modes()
        self.assertIsInstance(modes, list)
        for mode in modes:
            self.assertIn(mode, ALL_ROUTE_TYPES)
        for r_type in [2, 3]:
            self.assertIn(r_type, modes)

    def test_homogenize_stops_table_with_other_db(self):
        # TODO: testing this would require creating temp databases or something, joining inmemory databases might be impossible?
        pass

    def test_get_weekly_extract_start_date(self):
        trip_counts_per_day = self.G.get_trip_counts_per_day()
        first_day = trip_counts_per_day['date'].min()
        last_day = trip_counts_per_day['date'].max()  # a monday not in reach
        # print(first_day, last_day)
        first_monday = self.G.get_weekly_extract_start_date(download_date_override=first_day)
        early_monday = self.G.get_weekly_extract_start_date(
            download_date_override=first_day + datetime.timedelta(days=10))
        end_monday = self.G.get_weekly_extract_start_date(download_date_override=last_day - datetime.timedelta(days=5))
        assert first_monday < early_monday < end_monday



import copy
import pyximport

pyximport.install()

from unittest import TestCase

from gtfspy.routing.label import LabelTime, LabelTimeWithBoardingsCount, merge_pareto_frontiers, \
    LabelVehLegCount, compute_pareto_front, compute_pareto_front_naive, LabelTimeAndRoute, LabelTimeBoardingsAndRoute


class TestLabelTime(TestCase):

    def test_dominates(self):
        label1 = LabelTime(departure_time=0, arrival_time_target=20)
        label2 = LabelTime(departure_time=1, arrival_time_target=10)
        label3 = LabelTime(departure_time=1, arrival_time_target=10)
        label4 = LabelTime(departure_time=1, arrival_time_target=11)
        label5 = LabelTime(departure_time=0, arrival_time_target=10)
        self.assertTrue(label2.dominates(label1))
        self.assertTrue(label2.dominates(label3))
        self.assertTrue(label2.dominates(label4))
        self.assertTrue(label2.dominates(label5))

    def test_duration(self):
        label1 = LabelTime(departure_time=0, arrival_time_target=20)
        self.assertEqual(20, label1.duration())

    def test_equal(self):
        label1 = LabelTime(departure_time=0, arrival_time_target=20)
        label2 = LabelTime(departure_time=0, arrival_time_target=20)
        label3 = LabelTime(departure_time=10, arrival_time_target=20)
        self.assertEqual(label1, label2)
        self.assertNotEqual(label1, label3)

    def test_sort(self):
        l1 = LabelTime(departure_time=1, arrival_time_target=1)
        l2 = LabelTime(0, 0)
        self.assertTrue(l1 > l2)
        self.assertTrue(l1 >= l2)
        self.assertFalse(l1 < l2)
        self.assertFalse(l1 <= l2)

        l1 = LabelTime(0, 0)
        l2 = LabelTime(0, 0)
        self.assertTrue(l1 == l2)
        self.assertTrue(l1 >= l2)
        self.assertTrue(l1 <= l2)
        self.assertFalse(l1 != l2)

        l1 = LabelTime(1, 0)
        l2 = LabelTime(1, 1)
        self.assertTrue(l1 > l2)
        self.assertFalse(l1 < l2)

        self.assertTrue(sorted([l1, l2])[0] == l2)

    def test_large_numbers_do_not_overflow(self):
        departure_time = 1475530980
        arrival_time = 1475530980
        label = LabelTime(
            departure_time=float(departure_time),
            arrival_time_target=float(arrival_time),
            first_leg_is_walk=False
        )
        self.assertEqual(departure_time, label.departure_time)
        self.assertEqual(arrival_time, label.arrival_time_target)


class TestLabelTimeAndVehLegCount(TestCase):

    def test_dominates_simple(self):
        label1 = LabelTimeWithBoardingsCount(departure_time=0, arrival_time_target=20, n_boardings=0, first_leg_is_walk=False)
        label2 = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=10, n_boardings=0, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label1))

    def test_does_not_dominate_same(self):
        label2 = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=10, n_boardings=0, first_leg_is_walk=False)
        label3 = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=10, n_boardings=0, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label3))

    def test_dominates_later_arrival_time(self):
        label2 = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=10, n_boardings=0, first_leg_is_walk=False)
        label4 = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=11, n_boardings=0, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label4))

    def test_dominates_earlier_departure_time(self):
        label2 = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=10, n_boardings=0, first_leg_is_walk=False)
        label5 = LabelTimeWithBoardingsCount(departure_time=0, arrival_time_target=10, n_boardings=0, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label5))

    def test_dominates_less_transfers(self):
        labela = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=10, n_boardings=1, first_leg_is_walk=False)
        labelb = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=10, n_boardings=0, first_leg_is_walk=False)
        self.assertTrue(labelb.dominates(labela))

    def test_dominates_less_transfers_different_travel_time(self):
        labela = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=9, n_boardings=1, first_leg_is_walk=False)
        labelb = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=10, n_boardings=0, first_leg_is_walk=False)
        self.assertFalse(labelb.dominates(labela))
        self.assertFalse(labela.dominates(labelb))


    def test_duration(self):
        label1 = LabelTime(departure_time=0, arrival_time_target=20, last_leg_is_walk=False)
        self.assertEqual(20, label1.duration())

    def test_sort(self):
        l1 = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=1, n_boardings=3, first_leg_is_walk=False)
        l2 = LabelTimeWithBoardingsCount(0, 0, 0, False)
        self.assertTrue(l1 > l2)
        self.assertTrue(l1 >= l2)
        self.assertFalse(l1 < l2)
        self.assertFalse(l1 <= l2)

        l1 = LabelTimeWithBoardingsCount(0, 0, 0, False)
        l2 = LabelTimeWithBoardingsCount(0, 0, 0, False)
        self.assertTrue(l1 == l2)
        self.assertTrue(l1 >= l2)
        self.assertTrue(l1 <= l2)
        self.assertFalse(l1 != l2)

        l1 = LabelTimeWithBoardingsCount(1, 0, 10, False)
        l2 = LabelTimeWithBoardingsCount(1, 1, 10, False)
        self.assertTrue(l1 > l2)
        self.assertFalse(l1 < l2)

        l1 = LabelTimeWithBoardingsCount(1, 1, 0, False)
        l2 = LabelTimeWithBoardingsCount(1, 1, 10, False)
        self.assertTrue(l1 > l2)
        self.assertFalse(l1 < l2)

        self.assertTrue(sorted([l1, l2])[0] == l2)


        l1 = LabelTimeWithBoardingsCount(1, 1, 10, True)
        l2 = LabelTimeWithBoardingsCount(1, 1, 10, False)
        self.assertTrue(l1 < l2)
        self.assertFalse(l1 > l2)


    def test_large_numbers_do_not_overflow(self):
        departure_time = 1475530980
        arrival_time = 1475530980
        label = LabelTimeWithBoardingsCount(
            departure_time=float(departure_time),
            arrival_time_target=float(arrival_time),
            n_boardings=0,
            first_leg_is_walk=False
        )
        self.assertEqual(departure_time, label.departure_time)
        self.assertEqual(arrival_time, label.arrival_time_target)



class TestLabelVehLegCount(TestCase):

    def test_dominates_simple(self):
        label1 = LabelVehLegCount(n_boardings=1)
        label2 = LabelVehLegCount(n_boardings=0)
        self.assertTrue(label2.dominates(label1))


    def test_sort(self):
        l1 = LabelVehLegCount(departure_time=1, n_boardings=3)
        l2 = LabelVehLegCount(departure_time=0, n_boardings=0)
        self.assertTrue(l2 > l1)
        self.assertTrue(l2 >= l1)
        self.assertFalse(l2 < l1)
        self.assertFalse(l2 <= l1)

        l1 = LabelVehLegCount(departure_time=0, n_boardings=0)
        l2 = LabelVehLegCount(departure_time=0, n_boardings=0)
        self.assertTrue(l1 == l2)
        self.assertTrue(l1 >= l2)
        self.assertTrue(l1 <= l2)
        self.assertFalse(l1 != l2)

        l1 = LabelVehLegCount(departure_time=0, n_boardings=1)
        l2 = LabelVehLegCount(departure_time=0, n_boardings=0)
        self.assertFalse(l1 > l2)
        self.assertTrue(l1 < l2)
        self.assertTrue(l1 <= l2)
        self.assertTrue(l1 != l2)

        self.assertTrue(sorted([l1, l2])[0] == l1)

    def test_pareto_frontier(self):
        pt3 = LabelVehLegCount(departure_time=5, n_boardings=0)
        pt2 = LabelVehLegCount(departure_time=6, n_boardings=1)
        pt1 = LabelVehLegCount(departure_time=7, n_boardings=2)
        labels = [pt1, pt2, pt3]
        self.assertEqual(1, len(compute_pareto_front(labels)))


class TestLabelTimeAndRoute(TestCase):

    def test_dominates_simple(self):
        label1 = LabelTimeAndRoute(departure_time=0, arrival_time_target=20, movement_duration=0, first_leg_is_walk=False)
        label2 = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label1))
        self.assertFalse(label1.dominates(label2))

    def test_does_not_dominate_same(self):
        label2 = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        label3 = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label3))
        self.assertTrue(label3.dominates(label2))

    def test_dominates_later_arrival_time(self):
        label2 = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        label4 = LabelTimeAndRoute(departure_time=1, arrival_time_target=11, movement_duration=0, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label4))
        self.assertFalse(label4.dominates(label2))

    def test_dominates_earlier_departure_time(self):
        label2 = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        label5 = LabelTimeAndRoute(departure_time=0, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label5))
        self.assertFalse(label5.dominates(label2))

    def test_dominates_less_movement_duration(self):
        labela = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        labelb = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1, first_leg_is_walk=False)
        self.assertFalse(labelb.dominates(labela))
        self.assertTrue(labela.dominates(labelb))

    def test_dominates_less_movement_duration_when_arrival_time_not_the_same(self):
        # a should dominate b as the travel time is shorter
        labela = LabelTimeAndRoute(departure_time=1, arrival_time_target=9, movement_duration=1, first_leg_is_walk=False)
        labelb = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        self.assertFalse(labelb.dominates(labela))
        self.assertTrue(labela.dominates(labelb))

    def test_dominates_less_movement_duration_when_departure_time_not_the_same(self):
        labela = LabelTimeAndRoute(departure_time=4, arrival_time_target=10, movement_duration=1, first_leg_is_walk=False)
        labelb = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, first_leg_is_walk=False)
        self.assertFalse(labelb.dominates(labela))
        self.assertTrue(labela.dominates(labelb))

    def test_dominates_ignoring_dep_time_finalization_less_movement_duration(self):
        labela = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0,
                                   first_leg_is_walk=False)
        labelb = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1,
                                   first_leg_is_walk=False)
        self.assertTrue(labelb.dominates_ignoring_dep_time_finalization(labela))
        self.assertTrue(labela.dominates_ignoring_dep_time_finalization(labelb))

    def test_dominates_ignoring_dep_time_finalization_arrival_time(self):
        labela = LabelTimeAndRoute(departure_time=1, arrival_time_target=9, movement_duration=1,
                                   first_leg_is_walk=False)
        labelb = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1,
                                   first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time_finalization(labela))
        self.assertTrue(labela.dominates_ignoring_dep_time_finalization(labelb))

    def test_dominates_ignoring_dep_time_less_movement_duration(self):
        labela = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0,
                                   first_leg_is_walk=False)
        labelb = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1,
                                   first_leg_is_walk=False)
        self.assertTrue(labelb.dominates_ignoring_dep_time(labela))
        self.assertTrue(labela.dominates_ignoring_dep_time(labelb))

    def test_dominates_ignoring_dep_time_arrival_time(self):
        labela = LabelTimeAndRoute(departure_time=1, arrival_time_target=9, movement_duration=1,
                                   first_leg_is_walk=False)
        labelb = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1,
                                   first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time(labela))
        self.assertTrue(labela.dominates_ignoring_dep_time(labelb))

    """
    def test_dominates_ignoring_dep_time_finalization_equal(self):
        labela = LabelTimeAndRoute(departure_time=11, arrival_time_target=10, movement_duration=1,
                                   first_leg_is_walk=False)
        labelb = LabelTimeAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1,
                                   first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time_finalization(labela))
        self.assertFalse(labela.dominates_ignoring_dep_time_finalization(labelb))
    """
    def test_duration(self):
        label1 = LabelTimeAndRoute(departure_time=0, arrival_time_target=20, movement_duration=1, first_leg_is_walk=False)
        self.assertEqual(20, label1.duration())

    def test_sort(self):
        l1 = LabelTimeAndRoute(departure_time=1, arrival_time_target=1, movement_duration=3, first_leg_is_walk=False)
        l2 = LabelTimeAndRoute(0, 0, 0, False)
        self.assertTrue(l1 > l2)
        self.assertTrue(l1 >= l2)
        self.assertFalse(l1 < l2)
        self.assertFalse(l1 <= l2)

        l1 = LabelTimeAndRoute(0, 0, 0, False)
        l2 = LabelTimeAndRoute(0, 0, 0, False)
        self.assertTrue(l1 == l2)
        self.assertTrue(l1 >= l2)
        self.assertTrue(l1 <= l2)
        self.assertFalse(l1 != l2)

        l1 = LabelTimeAndRoute(1, 0, 10, False)
        l2 = LabelTimeAndRoute(1, 1, 10, False)
        self.assertTrue(l1 > l2)
        self.assertFalse(l1 < l2)

        l1 = LabelTimeAndRoute(1, 1, 0, False)
        l2 = LabelTimeAndRoute(1, 1, 10, False)
        self.assertTrue(l1 > l2)
        self.assertFalse(l1 < l2)

        self.assertTrue(sorted([l1, l2])[0] == l2)

        l1 = LabelTimeAndRoute(1, 1, 10, True)
        l2 = LabelTimeAndRoute(1, 1, 10, False)
        self.assertTrue(l1 < l2)
        self.assertFalse(l1 > l2)

    def test_large_numbers_do_not_overflow(self):
        departure_time = 1475530980
        arrival_time = 1475530980
        label = LabelTimeAndRoute(
            departure_time=float(departure_time),
            arrival_time_target=float(arrival_time),
            movement_duration=0,
            first_leg_is_walk=False
        )
        self.assertEqual(departure_time, label.departure_time)
        self.assertEqual(arrival_time, label.arrival_time_target)


class TestLabelTimeBoardingsAndRoute(TestCase):

    def test_dominates_simple(self):
        label1 = LabelTimeBoardingsAndRoute(departure_time=0, arrival_time_target=20, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        label2 = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label1))
        self.assertFalse(label1.dominates(label2))

    def test_does_not_dominate_same(self):
        label2 = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        label3 = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label3))
        self.assertTrue(label3.dominates(label2))

    def test_dominates_later_arrival_time(self):
        label2 = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        label4 = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=11, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label4))
        self.assertFalse(label4.dominates(label2))

    def test_dominates_earlier_departure_time(self):
        label2 = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        label5 = LabelTimeBoardingsAndRoute(departure_time=0, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        self.assertTrue(label2.dominates(label5))
        self.assertFalse(label5.dominates(label2))

    def test_dominates_less_movement_duration(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1, n_boardings=1, first_leg_is_walk=False)
        self.assertFalse(labelb.dominates(labela))
        self.assertTrue(labela.dominates(labelb))

    def test_dominates_less_movement_duration_when_arrival_time_not_the_same(self):
        # a should dominate b as the travel time is shorter
        labela = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=9, movement_duration=1, n_boardings=1, first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        self.assertFalse(labelb.dominates(labela))
        self.assertTrue(labela.dominates(labelb))

    def test_dominates_less_movement_duration_when_departure_time_not_the_same(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=4, arrival_time_target=10, movement_duration=1, n_boardings=1, first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1, first_leg_is_walk=False)
        self.assertFalse(labelb.dominates(labela))
        self.assertTrue(labela.dominates(labelb))

    def test_dominates_ignoring_dep_time_finalization_less_movement_duration(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1,
                                   first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1, n_boardings=1,
                                   first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time_finalization(labela))
        self.assertTrue(labela.dominates_ignoring_dep_time_finalization(labelb))

    def test_dominates_ignoring_dep_time_finalization_arrival_time(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=9, movement_duration=1, n_boardings=1,
                                            first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1,
                                            first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time_finalization(labela))
        self.assertTrue(labela.dominates_ignoring_dep_time_finalization(labelb))

    def test_dominates_ignoring_dep_time_finalization_both_pareto(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=9, movement_duration=1, n_boardings=1,
                                            first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=0,
                                            first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time_finalization(labela))
        self.assertFalse(labela.dominates_ignoring_dep_time_finalization(labelb))

    def test_dominates_ignoring_dep_time_less_movement_duration(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=0, n_boardings=1,
                                            first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1, n_boardings=1,
                                            first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time(labela))
        self.assertTrue(labela.dominates_ignoring_dep_time(labelb))

    def test_dominates_ignoring_dep_time_arrival_time(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=9, movement_duration=1, n_boardings=1,
                                            first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1, n_boardings=1,
                                            first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time(labela))
        self.assertTrue(labela.dominates_ignoring_dep_time(labelb))

    def test_various_dominates(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=1481520618, arrival_time_target=1481521300, n_boardings=1,
                                            movement_duration=681, first_leg_is_walk=True)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1481520618, arrival_time_target=1481521215, n_boardings=1,
                                            movement_duration=597, first_leg_is_walk=True)
        self.assertTrue(labelb.dominates(labela))
        self.assertFalse(labela.dominates(labelb))

    """
    def test_dominates_ignoring_dep_time_finalization_equal(self):
        labela = LabelTimeBoardingsAndRoute(departure_time=11, arrival_time_target=10, movement_duration=1,
                                   first_leg_is_walk=False)
        labelb = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=10, movement_duration=1,
                                   first_leg_is_walk=False)
        self.assertFalse(labelb.dominates_ignoring_dep_time_finalization(labela))
        self.assertFalse(labela.dominates_ignoring_dep_time_finalization(labelb))
    """
    def test_duration(self):
        label1 = LabelTimeBoardingsAndRoute(departure_time=0, arrival_time_target=20, movement_duration=1,
                                            n_boardings=1, first_leg_is_walk=False)
        self.assertEqual(20, label1.duration())

    def test_sort(self):
        l1 = LabelTimeBoardingsAndRoute(departure_time=1, arrival_time_target=1, movement_duration=3,
                                        n_boardings=1,  first_leg_is_walk=False)
        l2 = LabelTimeBoardingsAndRoute(0, 0, 0, 1, False)
        self.assertTrue(l1 > l2)
        self.assertTrue(l1 >= l2)
        self.assertFalse(l1 < l2)
        self.assertFalse(l1 <= l2)

        l1 = LabelTimeBoardingsAndRoute(0, 0, 0,  1, False)
        l2 = LabelTimeBoardingsAndRoute(0, 0, 0,  1, False)
        self.assertTrue(l1 == l2)
        self.assertTrue(l1 >= l2)
        self.assertTrue(l1 <= l2)
        self.assertFalse(l1 != l2)

        l1 = LabelTimeBoardingsAndRoute(1, 0, 10, 1, False)
        l2 = LabelTimeBoardingsAndRoute(1, 1, 10, 1, False)
        self.assertTrue(l1 > l2)
        self.assertFalse(l1 < l2)

        l1 = LabelTimeBoardingsAndRoute(1, 1, 0, 1, False)
        l2 = LabelTimeBoardingsAndRoute(1, 1, 10, 1, False)
        self.assertTrue(l1 > l2)
        self.assertFalse(l1 < l2)

        self.assertTrue(sorted([l1, l2])[0] == l2)

        l1 = LabelTimeBoardingsAndRoute(1, 1, 10, 1, True)
        l2 = LabelTimeBoardingsAndRoute(1, 1, 10, 1, False)
        self.assertTrue(l1 < l2)
        self.assertFalse(l1 > l2)

    def test_large_numbers_do_not_overflow(self):
        departure_time = 1475530980
        arrival_time = 1475530980
        label = LabelTimeBoardingsAndRoute(
            departure_time=float(departure_time),
            arrival_time_target=float(arrival_time),
            movement_duration=0,
            n_boardings=1,
            first_leg_is_walk=False
        )
        self.assertEqual(departure_time, label.departure_time)
        self.assertEqual(arrival_time, label.arrival_time_target)





class TestParetoFrontier(TestCase):

    def test_compute_pareto_front_all_include(self):
        label_a = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=2, n_boardings=0, first_leg_is_walk=False)
        label_b = LabelTimeWithBoardingsCount(departure_time=2, arrival_time_target=3, n_boardings=0, first_leg_is_walk=False)
        label_c = LabelTimeWithBoardingsCount(departure_time=3, arrival_time_target=4, n_boardings=0, first_leg_is_walk=False)
        label_d = LabelTimeWithBoardingsCount(departure_time=4, arrival_time_target=5, n_boardings=0, first_leg_is_walk=False)
        labels = [label_a, label_b, label_c, label_d]
        self.assertEqual(4, len(compute_pareto_front(labels)))

    def test_one_dominates_all(self):
        label_a = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=12, n_boardings=0, first_leg_is_walk=False)
        label_b = LabelTimeWithBoardingsCount(departure_time=2, arrival_time_target=13, n_boardings=0, first_leg_is_walk=False)
        label_c = LabelTimeWithBoardingsCount(departure_time=3, arrival_time_target=14, n_boardings=0, first_leg_is_walk=False)
        label_d = LabelTimeWithBoardingsCount(departure_time=4, arrival_time_target=5, n_boardings=0, first_leg_is_walk=False)
        labels = [label_a, label_b, label_c, label_d]
        pareto_front = compute_pareto_front(labels)
        self.assertEqual(1, len(pareto_front))
        self.assertEqual(label_d, pareto_front[0])

    def test_empty(self):
        labels = []
        self.assertEqual(0, len(compute_pareto_front(labels)))

    def test_some_are_optimal_some_are_not(self):
        label_a = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=2, n_boardings=1, first_leg_is_walk=False)  # optimal
        label_b = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=5, n_boardings=0, first_leg_is_walk=False)  # label_d dominates
        label_c = LabelTimeWithBoardingsCount(departure_time=3, arrival_time_target=4, n_boardings=1, first_leg_is_walk=False)  # optimal
        label_d = LabelTimeWithBoardingsCount(departure_time=4, arrival_time_target=5, n_boardings=0, first_leg_is_walk=False)  # optimal
        labels = [label_a, label_b, label_c, label_d]

        pareto_front = compute_pareto_front(labels)
        self.assertEqual(3, len(pareto_front))
        self.assertNotIn(label_b, pareto_front)

    def test_merge_pareto_frontiers(self):
        label_a = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=2, n_boardings=1, first_leg_is_walk=False)  # optimal
        label_b = LabelTimeWithBoardingsCount(departure_time=1, arrival_time_target=5, n_boardings=0, first_leg_is_walk=False)  # d dominates
        label_c = LabelTimeWithBoardingsCount(departure_time=3, arrival_time_target=4, n_boardings=1, first_leg_is_walk=False)  # optimal
        label_d = LabelTimeWithBoardingsCount(departure_time=4, arrival_time_target=5, n_boardings=0, first_leg_is_walk=False)  # optimal
        front_1 = [label_a, label_b]
        front_2 = [label_c, label_d]

        pareto_front = merge_pareto_frontiers(front_1, front_2)
        self.assertEqual(3, len(pareto_front))
        self.assertNotIn(label_b, pareto_front)

    def test_merge_pareto_frontiers_empty(self):
        pareto_front = merge_pareto_frontiers([], [])
        self.assertEqual(0, len(pareto_front))

    def test_compute_pareto_front_smart(self):
        labels = []
        for n in [1, 2, 10]: #, 500]:
            for dep_time in range(0, n):
                for n_veh_legs in range(2):
                    for arr_time in range(dep_time, dep_time  + 10):
                        label = LabelTimeWithBoardingsCount(dep_time, arr_time - n_veh_legs, n_veh_legs, False)
                        labels.append(label)
            import random
            random.shuffle(labels)
            labels_copy = copy.deepcopy(labels)
            pareto_optimal_labels = compute_pareto_front_naive(labels)
            self.assertEqual(len(pareto_optimal_labels), n * 2)
            pareto_optimal_labels = compute_pareto_front(labels_copy)
            self.assertEqual(len(pareto_optimal_labels), n * 2)

    def test_compute_pareto_front_smart_randomized(self):
        import random
        for i in range(10):
            labels = [LabelTimeWithBoardingsCount(random.randint(0, 1000), random.randint(0, 1000), random.randint(0, 10), 0)
                      for _ in range(1000)]
            pareto_optimal_labels_old = compute_pareto_front_naive(labels)
            pareto_optimal_labels_smart = compute_pareto_front(labels)
            self.assertEqual(len(pareto_optimal_labels_old), len(pareto_optimal_labels_smart))




from gtfspy.import_loaders.table_loader import TableLoader, decode_six


class StopLoader(TableLoader):
    # This class is documented to explain what it does, others are not.
    # Metadata needed to create table.  GTFS filename, table name, and
    # the CREATE TABLE syntax (last part only).
    fname = 'stops.txt'
    table = 'stops'
    tabledef = '''(stop_I INTEGER PRIMARY KEY, stop_id TEXT UNIQUE NOT NULL, code TEXT, name TEXT, desc TEXT, lat REAL, lon REAL, parent_I INT, location_type INT, wheelchair_boarding BOOL, self_or_parent_I INT)'''

    def gen_rows(self, readers, prefixes):
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                # and transform the "row" dictionary into a new
                # dictionary, which is yielded.  There can be different
                # transformations here, as needed.
                yield dict(
                    stop_id       = prefix + decode_six(row['stop_id']),
                    code          = decode_six(row['stop_code']) if 'stop_code' in row else None,
                    name          = decode_six(row['stop_name']),
                    desc          = decode_six(row['stop_desc']) if 'stop_desc' in row else None,
                    lat           = float(row['stop_lat']),
                    lon           = float(row['stop_lon']),
                    _parent_id    = prefix + decode_six(row['parent_station']) if row.get('parent_station','') else None,
                    location_type = int(row['location_type']) if row.get('location_type') else None,
                    wheelchair_boarding = int(row['wheelchair_boarding']) if row.get('wheelchair_boarding','') else None,
                )

    def post_import(self, cur):
        # if parent_id, set  also parent_I:
        # :_parent_id stands for a named parameter _parent_id
        # inputted through a dictionary in cur.executemany
        stmt = ('UPDATE %s SET parent_I=CASE WHEN (:_parent_id IS NOT "") THEN '
                '(SELECT stop_I FROM %s WHERE stop_id=:_parent_id) END '
                'WHERE stop_id=:stop_id') % (self.table, self.table)
        if self.exists():
            cur.executemany(stmt, self.gen_rows0())
        stmt = 'UPDATE %s ' \
               'SET self_or_parent_I=coalesce(parent_I, stop_I)' % self.table
        cur.execute(stmt)

    def index(self, cur):
        # Make indexes/ views as needed.
        #cur.execute('CREATE INDEX IF NOT EXISTS idx_stop_sid ON stop (stop_id)')
        #cur.execute('CREATE INDEX IF NOT EXISTS idx_stops_pid_sid ON stops (parent_id, stop_I)')
        #conn.commit()
        pass


import os
import sqlite3
import numpy as np

from pandas import read_sql_query, DataFrame, Series
from gtfspy.gtfs import GTFS
from gtfspy.util import timeit
from gtfspy.routing.journey_data import attach_database


class JourneyDataAnalyzer:
    # TODO: Transfer stops
    # TODO: circuity/directness

    def __init__(self, journey_db_path, gtfs_path):
        assert os.path.isfile(journey_db_path)
        assert os.path.isfile(gtfs_path)
        self.conn = sqlite3.connect(journey_db_path)
        self.g = GTFS(gtfs_path)
        self.gtfs_path = gtfs_path
        self.conn = attach_database(self.conn, self.gtfs_path)

    def __del__(self):
        self.conn.close()

    def get_journey_legs_to_target(self, target, fastest_path=True, min_boardings=False, all_leg_sections=True,
                                                   ignore_walk=False, diff_threshold=None, diff_path=None):
        """
        Returns a dataframe of aggregated sections from source nodes to target. The returned sections are either
        transfer point to transfer point or stop to stop. In a before after setting, the results can be filtered based
        on values in a difference db.
        :param target:
        :param fastest_path:
        :param min_boardings:
        :param all_leg_sections:
        :param ignore_walk:
        :param diff_threshold:
        :param diff_path:
        :return:
        """
        assert not (fastest_path and min_boardings)
        if min_boardings:
            raise NotImplementedError
        if all_leg_sections and diff_threshold:
            raise NotImplementedError

        added_constraints = ""
        add_diff = ""
        if fastest_path:
            added_constraints += " AND journeys.pre_journey_wait_fp>=0"
        if ignore_walk:
            added_constraints += " AND legs.trip_I >= 0"
        if diff_path and diff_threshold:
            self.conn = attach_database(self.conn, diff_path, name="diff")
            add_diff = ", diff.diff_temporal_distance"
            added_constraints += " AND abs(diff_temporal_distance.diff_mean) >= %s " \
                                 "AND diff_temporal_distance.from_stop_I = journeys.from_stop_I " \
                                 "AND diff_temporal_distance.to_stop_I = journeys.to_stop_I" % (diff_threshold,)

        if all_leg_sections:
            df = self._get_journey_legs_to_target_with_all_sections(target, added_constraints)
        else:
            query = """SELECT from_stop_I, to_stop_I, coalesce(type, -1) AS type,
                         count(*) AS n_trips
                         FROM
                         (SELECT legs.* FROM legs, journeys %s
                         WHERE journeys.journey_id = legs.journey_id AND journeys.to_stop_I = %s %s) q1
                         LEFT JOIN (SELECT * FROM other.trips, other.routes WHERE trips.route_I = routes.route_I) q2
                         ON q1.trip_I = q2.trip_I
                         GROUP BY from_stop_I, to_stop_I, type""" % (add_diff, str(target), added_constraints)
            df = read_sql_query(query, self.conn)

        return df

    def _get_journey_legs_to_target_with_all_sections(self, target, added_constraint):
        def gen_pairs(stop_lists):
            for stop_list in stop_lists:
                prev_stop = None
                stop_pair_list = []
                for stop in stop_list:
                    if prev_stop:
                        stop_pair_list.append((int(prev_stop), int(stop)))
                    prev_stop = stop
                yield stop_pair_list

        query = """SELECT leg_stops, coalesce(type, -1) AS type, count(*) AS n_trips FROM
                   (SELECT legs.* FROM legs, journeys
                   WHERE journeys.journey_id = legs.journey_id AND journeys.to_stop_I = %s %s) q1
                     LEFT JOIN (SELECT * FROM other.trips, other.routes WHERE trips.route_I = routes.route_I) q2
                     ON q1.trip_I = q2.trip_I
                     GROUP BY leg_stops, type""" % (str(target), added_constraint)
        orig_df = read_sql_query(query, self.conn)

        df = DataFrame([x for x in gen_pairs(orig_df.leg_stops.str.split(',').tolist())],
                          index=[orig_df.type, orig_df.n_trips]).stack()

        df = df.reset_index()
        df = df.rename(columns={0: "stop_tuple"})
        df[['from_stop_I', 'to_stop_I']] = df['stop_tuple'].apply(Series)

        df = df.groupby(['from_stop_I', 'to_stop_I', 'type']).agg({'n_trips': [np.sum]})
        df = df.reset_index()
        df.columns = df.columns.droplevel(1)
        df_to_return = df[['from_stop_I', 'to_stop_I', 'type', 'n_trips']]

        return df_to_return

    def get_origin_target_journey_legs(self, origin, target, start_time=None, end_time=None, fastest_path=True, min_boardings=False,
                                       ignore_walk=False, add_coordinates=True):

        assert not (fastest_path and min_boardings)
        if min_boardings:
            raise NotImplementedError

        added_constraints = ""
        if fastest_path:
            added_constraints += " AND journeys.pre_journey_wait_fp>=0"
        if ignore_walk:
            added_constraints += " AND legs.trip_I >= 0"
        if start_time:
            added_constraints += " AND journeys.departure_time>= %s" % start_time
        if start_time:
            added_constraints += " AND journeys.departure_time< %s" % end_time

        query = """SELECT from_stop_I, to_stop_I, coalesce(type, -1) AS type,
                     count(*) AS n_trips, group_concat(dep_time) AS dep_times
                     FROM
                     (SELECT legs.*, journeys.departure_time as dep_time FROM legs, journeys
                     WHERE journeys.journey_id = legs.journey_id AND journeys.from_stop_I = %s 
                     AND journeys.to_stop_I = %s %s
                     ORDER BY dep_time
                     ) q1
                     LEFT JOIN (SELECT * FROM other.trips, other.routes WHERE trips.route_I = routes.route_I) q2
                     ON q1.trip_I = q2.trip_I
                     GROUP BY from_stop_I, to_stop_I, type""" % (str(origin), str(target), added_constraints)
        df = read_sql_query(query, self.conn)
        if add_coordinates:
            df = self.g.add_coordinates_to_df(df, join_column="from_stop_I", lat_name="from_lat", lon_name="from_lon")
            df = self.g.add_coordinates_to_df(df, join_column="to_stop_I", lat_name="to_lat", lon_name="to_lon")
        return df

    def get_journey_routes_not_in_other_db(self, target, other_journey_conn, fastest_path=True, min_boardings=False, all_leg_sections=True,
                                           ignore_walk=False, diff_threshold=None, diff_path=None):
        name = "ojdb"
        added_constraints = ""
        if fastest_path:
            added_constraints += " AND journeys.pre_journey_wait_fp>=0"
        if ignore_walk:
            added_constraints += " AND legs.trip_I >= 0"

        query = """SELECT from_stop_I, to_stop_I, coalesce(type, -1) AS type, route FROM 
                    (SELECT legs.*, route FROM journeys, legs WHERE legs.journey_id=journeys.journey_id AND journeys.to_stop_I = %s %s) q1
                    LEFT JOIN 
                    (SELECT * FROM other.trips, other.routes WHERE trips.route_I = routes.route_I) q2
                    ON q1.trip_I = q2.trip_I
                    """ % (str(target), added_constraints)

        df = read_sql_query(query, self.conn)

        routes = other_journey_conn.execute("SELECT DISTINCT route FROM journeys WHERE to_stop_I = %s" % (str(target),)).fetchall()
        routes = [x[0] for x in routes]
        other_set = set(routes)
        these_routes = self.conn.execute("SELECT DISTINCT route FROM journeys WHERE to_stop_I = %s" % (str(target),)).fetchall()
        these_routes = [x[0] for x in these_routes]
        this_set = set(these_routes)
        print("n unique routes for this db: ", len(this_set))
        print("n unique routes for other db: ", len(other_set))

        union = other_set | this_set
        intersection = other_set & this_set
        print("n unique routes", len(union))
        print("n common routes", len(intersection))

        df = df.loc[~df['route'].isin(routes)]
        df = df[["from_stop_I", "to_stop_I", "type"]]
        df = DataFrame({"n_trips": df.groupby(["from_stop_I", "to_stop_I", "type"]).size()}).reset_index()
        return df

    def journey_alternatives_per_stop_pair(self, target, start_time, end_time):
        query = """SELECT from_stop_I, to_stop_I, ifnull(1.0*sum(n_sq)/(sum(n_trips)*(sum(n_trips)-1)), 1) AS simpson,
                    sum(n_trips) AS n_trips, count(*) AS n_routes FROM 
                    (SELECT from_stop_I, to_stop_I, count(*) AS n_trips, count(*)*(count(*)-1) AS n_sq 
                    FROM journeys
                    WHERE pre_journey_wait_fp > 0 AND to_stop_I = %s AND departure_time >= %s AND departure_time <= %s
                    GROUP BY route) sq1
                    GROUP BY from_stop_I, to_stop_I""" % (target, start_time, end_time)
        df = read_sql_query(query, self.conn)
        df = self.g.add_coordinates_to_df(df, join_column="from_stop_I")

        return df

    def journey_alternative_data_time_weighted(self, target, start_time, end_time):
        query = """SELECT sum(p*p) AS simpson, sum(n_trips) AS n_trips, count(*) AS n_routes, from_stop_I, to_stop_I FROM
                    (SELECT 1.0*sum(pre_journey_wait_fp)/total_time AS p, count(*) AS n_trips, route, 
                    journeys.from_stop_I, journeys.to_stop_I FROM journeys,
                    (SELECT sum(pre_journey_wait_fp) AS total_time, from_stop_I, to_stop_I FROM journeys
                    WHERE departure_time >= %s AND departure_time <= %s
                    GROUP BY from_stop_I, to_stop_I) sq1
                    WHERE pre_journey_wait_fp > 0 AND sq1.to_stop_I=journeys.to_stop_I AND departure_time >= %s 
                    AND departure_time <= %s AND journeys.to_stop_I = %s AND sq1.from_stop_I = journeys.from_stop_I 
                    GROUP BY route) sq2
                    GROUP BY from_stop_I, to_stop_I""" % (start_time, end_time, start_time, end_time, target)
        df = read_sql_query(query, self.conn)
        df = self.g.add_coordinates_to_df(df, join_column="from_stop_I")

        return df

    def _add_to_from_coordinates_to_df(self, df):
        df = self.g.add_coordinates_to_df(df, join_column="from_stop_I", lat_name="from_lat", lon_name="from_lon")
        df = self.g.add_coordinates_to_df(df, join_column="to_stop_I", lat_name="to_lat", lon_name="to_lon")
        return df

    def get_upstream_stops(self, target, stop):
        query = """SELECT stops.* FROM other.stops, 
                    (SELECT journeys.from_stop_I AS stop_I FROM journeys, legs 
                    WHERE journeys.journey_id=legs.journey_id AND legs.from_stop_I = %s AND journeys.to_stop_I = %s AND pre_journey_wait_fp >= 0
                    GROUP BY journeys.from_stop_I) q1
                    WHERE stops.stop_I = q1.stop_I""" % (stop, target)
        df = read_sql_query(query, self.conn)
        return df

    def get_upstream_stops_ratio(self, target, trough_stops, ratio):
        """
        Selects the stops for which the ratio or higher proportion of trips to the target passes trough a set of trough stops
        :param target: target of trips
        :param trough_stops: stops where the selected trips are passing trough
        :param ratio: threshold for inclusion
        :return:
        """
        if isinstance(trough_stops, list):
            trough_stops = ",".join(trough_stops)
        query = """SELECT stops.* FROM other.stops, 
                    (SELECT q2.from_stop_I AS stop_I FROM 
                    (SELECT journeys.from_stop_I, count(*) AS n_total FROM journeys
                    WHERE journeys.to_stop_I = {target} 
                    GROUP BY from_stop_I) q1,
                    (SELECT journeys.from_stop_I, count(*) AS n_trough FROM journeys, legs 
                    WHERE journeys.journey_id=legs.journey_id AND legs.from_stop_I IN ({trough_stops}) AND journeys.to_stop_I = {target}
                    GROUP BY journeys.from_stop_I) q2
                    WHERE q1.from_stop_I = q2.from_stop_I AND n_trough/(n_total*1.0) >= {ratio}) q1
                    WHERE stops.stop_I = q1.stop_I""".format(target=target, trough_stops=trough_stops, ratio=ratio)
        df = read_sql_query(query, self.conn)
        return df

    def passing_journeys_per_stop(self):
        """

        :return:
        """
        pass
    @timeit
    def journeys_per_section(self, fastest_path=False, time_weighted=False):
        """

        :return:
        """
        pass

    def n_departure_stop_alternatives(self):
        """

        :return:
        """
        pass

    def aggregate_in_vehicle_times(self, per_mode):
        pass

    def aggregate_in_vehicle_distances(self, per_mode):
        pass

    def aggregate_walking_times(self):
        pass

    def aggregate_walking_distance(self):
        pass

    @timeit
    def get_transfer_stops(self, group_by_routes=False):
        pass

    @timeit
    def get_transfer_walks(self, group_by_routes=False):
        pass

    def get_journey_distance(self):
        pass

    def get_journey_time(self):
        """
        (using the connection objects)
        :return:
        """
        pass

    def get_journey_time_per_mode(self, modes=None):
        """

        :param modes: return these
        :return:
        """
        pass

    def get_walking_time(self):
        pass




from .table_loader import TableLoader, decode_six


class CalendarLoader(TableLoader):
    fname = 'calendar.txt'
    table = 'calendar'
    tabledef = '(service_I INTEGER PRIMARY KEY, service_id TEXT UNIQUE NOT NULL, m INT, t INT, w INT, th INT, f INT, s INT, su INT, start_date TEXT, end_date TEXT)'
    copy_where = ("WHERE  date({start_ut}, 'unixepoch', 'localtime') < end_date "
                  "AND  start_date < date({end_ut}, 'unixepoch', 'localtime')")

    # service_id,monday,tuesday,wednesday,thursday,friday,saturday,sunday,start_date,end_date
    # 1001_20150810_20151014_Ke,0,0,1,0,0,0,0,20150810,20151014
    def gen_rows(self, readers, prefixes):
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                # print row
                start = row['start_date']
                end = row['end_date']
                yield dict(
                    service_id    = prefix + decode_six(row['service_id']),
                    m             = int(row['monday']),
                    t             = int(row['tuesday']),
                    w             = int(row['wednesday']),
                    th            = int(row['thursday']),
                    f             = int(row['friday']),
                    s             = int(row['saturday']),
                    su            = int(row['sunday']),
                    start_date    = '%s-%s-%s' % (start[:4], start[4:6], start[6:8]),
                    end_date      = '%s-%s-%s' % (end[:4], end[4:6], end[6:8]),
                )

    @classmethod
    def index(cls, cur):
        # cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_svid ON calendar (service_id)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_s_e ON calendar (start_date, end_date)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_m  ON calendar (m)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_t  ON calendar (t)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_w  ON calendar (w)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_th ON calendar (th)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_f  ON calendar (f)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_s  ON calendar (s)')
        cur.execute('CREATE INDEX IF NOT EXISTS idx_calendar_su ON calendar (su)')


import codecs
import csv
import os
import sys
import zipfile

from six import string_types

from gtfspy import util


class TableLoader(object):
    """Generic GTFS table loader.

    This is a generic table loader that can load any other GTFS table.
    The reason that this is a class is so that it can be subclassed
    and modified to easily import each other table without writing a
    bunch of redundant code.

    This class is just instantiated, and it does its stuff, and then
    it is destroyed.
    """
    mode = 'all'  # None or 'import' or 'index'.  "None" does everything.

    # The following properties need to be defined in a subclass. Examples here.
    # fname = 'routes.txt'
    # table = 'route'
    # tabledef = '(route TEXT PRIMARY KEY,
    #               agency_id TEXT,
    #               name TEXT,
    #               long_name TEXT,
    #               desc TEXT,
    #               type INT,
    #               url TEXT)'
    # Finally, a subclass needs to define these methods:
    # def gen_rows(self, reader):
    # def index(self):
    extra_keys = []
    extra_values = []
    is_zipfile = False
    table = ""  # e.g. stops for StopLoader

    def __init__(self, gtfssource=None, print_progress=True):
        """
        Parameters
        ----------
        gtfssource: str, dict, list
            str: path to GTFS directory or zipfile
            dict:
                dictionary of files to use to as the GTFS files.  This
                is mainly useful for testing, not for any normal use of
                GTFS.  For example, to provide an agency.txt file,
                do this:
                    d = {'agency.txt':
                             'agency_id, agency_name, agency_timezone,agency_url\n' \
                             '555,CompNet,Europe/Lala,http://x'
                          }
                Of course you probably wouldn't want all the string data
                inline like that. You can provide the data as a string or
                a file-like object (with a .read() attribute and line
                iteration).
            list: a list of the above elements to import (i.e. "merge") multiple GTFS feeds to the same database

        print_progress: boolean
            whether to print progress of the
        """
        if isinstance(gtfssource, string_types + (dict,)):
            _gtfs_sources = [gtfssource]
        else:
            assert isinstance(gtfssource, list)
            _gtfs_sources = gtfssource

        # whether to print progress of the import
        self.print_progress = print_progress

        self.gtfs_sources = []
        # map sources to "real"
        for source in _gtfs_sources:
            # print(source)
            # Deal with the case that gtfspath is actually a dict.
            if isinstance(source, dict):
                self.gtfs_sources.append(source)
            # zipfile input.
            # Warning: this keeps an open file handle as
            # long as this object exists, and it is duplicated across all
            # loaders. Maybe open it in the caller?
            elif isinstance(source, string_types):
                if os.path.isdir(source):
                    self.gtfs_sources.append(source)
                else:
                    z = zipfile.ZipFile(source, mode='r')
                    zip_commonprefix = os.path.commonprefix(z.namelist())
                    zip_source_datum = {
                        "zipfile": source,
                        "zip_commonprefix": zip_commonprefix
                    }
                    self.gtfs_sources.append(zip_source_datum)

    # Methods that should be implemented by inheriting classes
    # when necessary.
    #
    #def post_import(self, cur):
    #def index(self, cur):

    # Methods for these classes:

    def exists(self):
        """Does this GTFS contain this file? (file specified by the class)"""
        # If file exists in one of the sources, return True
        return any(self.exists_by_source())

    def exists_by_source(self):
        """Does this GTFS contain this file? (file specified by the class)"""
        exists_list = []
        for source in self.gtfs_sources:
            if isinstance(source, dict):
                # source can now be either a dict or a zipfile
                if self.fname in source:
                    if source[self.fname]:
                        exists_list.append(True)
                        continue
            # Handle zipfiles specially
            if "zipfile" in source:
                try:
                    Z = zipfile.ZipFile(source['zipfile'], mode='r')
                    Z.getinfo(os.path.join(source['zip_commonprefix'], self.fname))
                    exists_list.append(True)
                    continue
                # File does not exist in the zip archive
                except KeyError:
                    print(self.fname, ' missing in ', source)
                    exists_list.append(False)
                    continue
            # Normal filename
            elif isinstance(source, string_types):
                if os.path.exists(os.path.join(source, self.fname)):
                    exists_list.append(True)
                    continue
            exists_list.append(False)
        # the "file" was not found in any of the sources, return false
        return exists_list

    def assert_exists_if_required(self):
        REQUIRED_FILES_GTFS = ["agency.txt", "stops.txt", "routes.txt", "trips.txt", "stop_times.txt"]
        if self.fname in REQUIRED_FILES_GTFS:
            for gtfs_source, exists in zip(self.gtfs_sources, self.exists_by_source()):
                if not exists:
                    raise AssertionError(self.fname + " does not exist in the provided gtfs_source")

    def gen_rows0(self):
        """Iterate through all rows in all files file.

        The file is specified by the class - there is one class per
        table.  This opens the file, does basic sanitaition, and
        iterates over all rows as dictionaries, converted by
        csv.DictReader.  This function opens files from both .zip and
        raw directories.  The actual logic of converting all data to
        Python is done in the .gen_rows() method which must be
        defined in each subclass.
        """
        return self.gen_rows(*(self._get_csv_reader_generators()))

    def _get_csv_reader_generators(self):
        # This is a generator function that we use for importing.  It
        # makes a CSV reader that returns dictionaries, and passes
        # that to self.gen_rows that transform those CSV dictionaries
        # into the right form for importing into SQLite.  It is worth
        # pointing out that dictionaries are used everywhere here to
        # not have to depend on the particular ordering of fields, and
        # to make it easier to add more fields in the future.
        def _iter_file(file_obj):
            # This hack removes the BOM from the start of any
            # line.
            for line in file_obj:
                yield line.lstrip(codecs.BOM_UTF8.decode("utf-8"))

        fs = []
        for source in self.gtfs_sources:
            f = []
            # Handle manually overridden files.
            if isinstance(source, dict):
                # source can now be either a dict or a zipfile
                if self.fname in source:
                    data_obj = source[self.fname]
                    if isinstance(data_obj, string_types):
                        f = data_obj.split("\n")
                    elif hasattr(data_obj, "read"):
                        # file-like object: use it as-is.
                        f = data_obj
                elif "zipfile" in source:
                    try:
                        Z = zipfile.ZipFile(source['zipfile'], mode='r')
                        # print(Z.namelist())
                        f = util.zip_open(Z, os.path.join(source['zip_commonprefix'], self.fname))
                    except KeyError:
                        pass
            elif isinstance(source, string_types):
                # now source is a directory
                try:
                    f = open(os.path.join(source, self.fname))
                # except OSError as e:
                except IOError as e:
                    f = []
            fs.append(f)

        csv_readers = [csv.DictReader(_iter_file(f)) for f in fs]
        csv_reader_generators = []
        for i, csv_reader in enumerate(csv_readers):
            try:
                csv_reader.fieldnames = [x.strip() for x in csv_reader.fieldnames]
                # Make a generator that strips the values in all fields before
                # passing it on.  GTFS standard requires that there be no
                # spare whitespace, but not all do it.  The csv module has a
                # `skipinitialspace` option, but let's make sure that we strip
                # it from both sides.
                # The following results in a generator, the complicated
                csv_reader_stripped = (dict((k, (v.strip() if v is not None else None))  # v is not always a string
                                            for k, v in row.items())
                                       for row in csv_reader)
                csv_reader_generators.append(csv_reader_stripped)
            except TypeError as e:
                if "NoneType" in str(e):
                    print(self.fname + " missing from feed " + str(i))
                    csv_reader_generators.append(iter(()))
                    #raise e here will make every multifeed download with incompatible number of tables fail
                else:
                    raise e
        prefixes = [u"feed_{i}_".format(i=i) for i in range(len(csv_reader_generators))]

        if len(prefixes) == 1:
            # no prefix for a single source feed
            prefixes = [u""]
        return csv_reader_generators, prefixes

    def gen_rows(self, csv_readers, prefixes):
        # to be overridden by Inherited classes
        pass

    def create_table(self, conn):
        """Make table definitions"""
        # Make cursor
        cur = conn.cursor()
        # Drop table if it already exists, to be recreated.  This
        # could in the future abort if table already exists, and not
        # recreate it from scratch.
        #cur.execute('''DROP TABLE IF EXISTS %s'''%self.table)
        #conn.commit()
        if self.tabledef is None:
            return
        if not self.tabledef.startswith('CREATE'):
            # "normal" table creation.
            cur.execute('CREATE TABLE IF NOT EXISTS %s %s'
                        % (self.table, self.tabledef)
                        )
        else:
            # When tabledef contains the full CREATE statement (for
            # virtual tables).
            cur.execute(self.tabledef)
        conn.commit()

    def insert_data(self, conn):
        """Load data from GTFS file into database"""
        cur = conn.cursor()
        # This is a bit hackish.  It is annoying to have to write the
        # INSERT statement yourself and keep it up to date with the
        # table rows.  This gets the first row, figures out the field
        # names from that, and then makes an INSERT statement like
        # "INSERT INTO table (col1, col2, ...) VALUES (:col1, :col2,
        # ...)".  The ":col1" is sqlite syntax for named value.

        csv_reader_generators, prefixes = self._get_csv_reader_generators()
        for csv_reader, prefix in zip(csv_reader_generators, prefixes):
            try:
                row = next(iter(self.gen_rows([csv_reader], [prefix])))
                fields = row.keys()
            except StopIteration:
                # The file has *only* a header and no data.
                # next(iter()) yields StopIteration and we can't
                # proceed.  Since there is nothing to import, just continue the loop
                print("Not importing %s into %s for %s" % (self.fname, self.table, prefix))
                continue
            stmt = '''INSERT INTO %s (%s) VALUES (%s)''' % (
                self.table,
                (', '.join([x for x in fields if x[0] != '_'] + self.extra_keys)),
                (', '.join([":" + x for x in fields if x[0] != '_'] + self.extra_values))
            )

            # This does the actual insertions.  Passed the INSERT
            # statement and then an iterator over dictionaries.  Each
            # dictionary is inserted.
            if self.print_progress:
                print('Importing %s into %s for %s' % (self.fname, self.table, prefix))
            # the first row was consumed by fetching the fields
            # (this could be optimized)
            from itertools import chain
            rows = chain([row], self.gen_rows([csv_reader], [prefix]))
            cur.executemany(stmt, rows)
            conn.commit()

            # This was used for debugging the missing service_I:
            # if self.__class__.__name__ == 'TripLoader': # and False:
                # for i in self.gen_rows([new_csv_readers[i]], [prefix]):
                # print(stmt)
                # rows = cur.execute('SELECT agency_id, trips.service_id FROM agencies, routes, trips
            # LEFT JOIN calendar ON(calendar.service_id=trips.service_id)
            # WHERE trips.route_I = routes.route_I and routes.agency_I = agencies.agency_I and trips.service_I is NULL
            # GROUP BY trips.service_id, agency_id').fetchall()
                # rows = cur.execute('SELECT distinct trips.service_id FROM trips
            # LEFT JOIN calendar ON(calendar.service_id=trips.service_id) WHERE trips.service_I is NULL').fetchall()

                # print('trips, etc', [description[0] for description in cur.description])
                # for i, row in enumerate(rows):
                    # print(row)
                    #if i == 100:
                        #exit(0)

                # rows = cur.execute('SELECT distinct service_id FROM calendar').fetchall()
                # print('calendar_columns',[description[0] for description in cur.description])
                # for row in rows:
                    # print(row)

    def run_post_import(self, conn):
        if self.print_progress:
            print('Post-import %s into %s' % (self.fname, self.table))
        cur = conn.cursor()
        self.post_import(cur)
        conn.commit()

    def create_index(self, conn):
        if not hasattr(self, 'index'):
            return
        cur = conn.cursor()
        if self.print_progress:
            print('Indexing %s' % (self.table,))
        self.index(cur)
        conn.commit()

    def import_(self, conn):
        """Do the actual import. Copy data and store in connection object.

        This function:
        - Creates the tables
        - Imports data (using self.gen_rows)
        - Run any post_import hooks.
        - Creates any indexs
        - Does *not* run self.make_views - those must be done
          after all tables are loaded.
        """
        if self.print_progress:
            print('Beginning', self.__class__.__name__)
        # what is this mystical self._conn ?
        self._conn = conn

        self.create_table(conn)
        # This does insertions
        if self.mode in ('all', 'import') and self.fname and self.exists() and self.table not in ignore_tables:
            self.insert_data(conn)
        # This makes indexes in the DB.
        if self.mode in ('all', 'index') and hasattr(self, 'index'):
            self.create_index(conn)
        # Any post-processing to be done after the full import.
        if self.mode in ('all', 'import') and hasattr(self, 'post_import'):
            self.run_post_import(conn)
        # Commit it all
        conn.commit()

    @classmethod
    def make_views(cls, conn):
        """The make views should be run after all tables imported."""
        pass

    copy_where = ''

    @classmethod
    def copy(cls, conn, **where):
        """Copy data from one table to another while filtering data at the same time

        Parameters
        ----------
        conn: sqlite3 DB connection.  It must have a second database
            attached as "other".
        **where : keyword arguments
            specifying (start_ut and end_ut for filtering, see the copy_where clause in the subclasses)
        """
        cur = conn.cursor()
        if where and cls.copy_where:
            copy_where = cls.copy_where.format(**where)
            # print(copy_where)
        else:
            copy_where = ''
        cur.execute('INSERT INTO %s '
                    'SELECT * FROM source.%s %s' % (cls.table, cls.table, copy_where))

    @classmethod
    def post_import_round2(cls, conn):
        pass


ignore_tables = set()


def decode_six(string):
    version = sys.version_info[0]
    if version == 2:
        return string.decode('utf-8')
    else:
        assert(isinstance(string, str))
        return string


import unittest

import networkx

from gtfspy.routing.connection_scan_profile import ConnectionScanProfiler
from gtfspy.routing.connection import Connection
from gtfspy.routing.label import LabelTimeSimple


# noinspection PyAttributeOutsideInit

class ConnectionScanProfileTest(unittest.TestCase):

    def setUp(self):
        event_list_raw_data = [
            (2, 4, 40, 50, "trip_6", 1),
            (1, 3, 32, 40, "trip_5", 1),
            (3, 4, 32, 35, "trip_4", 1),
            (2, 3, 25, 30, "trip_3", 1),
            (1, 2, 10, 20, "trip_2", 1),
            (0, 1, 0, 10, "trip_1", 1)
        ]
        self.transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        self.walk_network = networkx.Graph()
        self.walk_network.add_edge(1, 2, {"d_walk": 20})
        self.walk_network.add_edge(3, 4, {"d_walk": 15})
        self.walk_speed = 1
        self.source_stop = 1
        self.target_stop = 4
        self.transfer_margin = 0
        self.start_time = 0
        self.end_time = 50

    def test_basics(self):
        csa_profile = ConnectionScanProfiler(self.transit_connections, self.target_stop,
                                             self.start_time, self.end_time, self.transfer_margin,
                                             self.walk_network, self.walk_speed)
        csa_profile.run()

        stop_3_pareto_tuples = csa_profile.stop_profiles[3].get_final_optimal_labels()
        self.assertEqual(len(stop_3_pareto_tuples), 1)
        self.assertIn(LabelTimeSimple(32, 35), stop_3_pareto_tuples)

        stop_2_pareto_tuples = csa_profile.stop_profiles[2].get_final_optimal_labels()
        self.assertEqual(len(stop_2_pareto_tuples), 2)
        self.assertIn(LabelTimeSimple(40, 50), stop_2_pareto_tuples)
        self.assertIn(LabelTimeSimple(25, 35), stop_2_pareto_tuples)

        source_stop_profile = csa_profile.stop_profiles[self.source_stop]
        source_stop_pareto_optimal_tuples = source_stop_profile.get_final_optimal_labels()

        pareto_tuples = list()
        pareto_tuples.append(LabelTimeSimple(departure_time=10, arrival_time_target=35))
        pareto_tuples.append(LabelTimeSimple(departure_time=20, arrival_time_target=50))
        pareto_tuples.append(LabelTimeSimple(departure_time=32, arrival_time_target=55))

        self._assert_pareto_labels_equal(
            pareto_tuples,
            source_stop_pareto_optimal_tuples
        )

    def test_wrong_event_data_ordering(self):
        event_list_wrong_ordering = [
            (0, 1, 0, 10, "trip_1", 1),
            (1, 2, 10, 20, "trip_2", 1),
            (2, 3, 25, 30, "trip_3", 1),
            (3, 4, 32, 35, "trip_4", 1),
            (1, 3, 32, 40, "trip_5", 1),
            (2, 4, 40, 50, "trip_5", 1)
        ]
        csa_profile = ConnectionScanProfiler(event_list_wrong_ordering, self.target_stop,
                                             self.start_time, self.end_time, self.transfer_margin,
                                             self.walk_network, self.walk_speed)
        self.assertRaises(AssertionError, csa_profile.run)

    def test_simple(self):
        event_list_raw_data = [
            (2, 4, 40, 50, "trip_5", 1),
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 20})
        walk_network.add_edge(3, 4, {"d_walk": 15})
        walk_speed = 1
        source_stop = 1
        target_stop = 4
        transfer_margin = 0
        start_time = 0
        end_time = 50

        pareto_tuples = list()
        pareto_tuples.append(LabelTimeSimple(departure_time=20, arrival_time_target=50))

        csa_profile = ConnectionScanProfiler(transit_connections, target_stop,
                                             start_time, end_time, transfer_margin,
                                             walk_network, walk_speed)
        csa_profile.run()
        source_stop_profile = csa_profile.stop_profiles[source_stop]
        source_stop_pareto_tuples = source_stop_profile.get_final_optimal_labels()

        self._assert_pareto_labels_equal(
            pareto_tuples,
            source_stop_pareto_tuples
        )

    def test_last_leg_is_walk(self):
        event_list_raw_data = [
            (0, 1, 0, 10, "trip_1", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 20})

        walk_speed = 1
        source_stop = 0
        target_stop = 2
        transfer_margin = 0
        start_time = 0
        end_time = 50
        pareto_tuples = list()
        pareto_tuples.append(LabelTimeSimple(departure_time=0, arrival_time_target=30))

        csa_profile = ConnectionScanProfiler(transit_connections, target_stop,
                                             start_time, end_time, transfer_margin,
                                             walk_network, walk_speed)
        csa_profile.run()
        found_tuples = csa_profile.stop_profiles[source_stop].get_final_optimal_labels()
        self._assert_pareto_labels_equal(found_tuples, pareto_tuples)

    def test_walk_is_faster_than_by_trip(self):
        event_list_raw_data = [
            (0, 1, 0, 10, "trip_1", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_speed = 2
        source_stop = 0
        target_stop = 1
        transfer_margin = 0
        start_time = 0
        end_time = 50

        walk_network = networkx.Graph()
        walk_network.add_edge(0, 1, {"d_walk": 1})
        csa_profile = ConnectionScanProfiler(transit_connections, target_stop,
                                             start_time, end_time, transfer_margin,
                                             walk_network, walk_speed)
        csa_profile.run()
        source_profile = csa_profile.stop_profiles[source_stop]
        self.assertEqual(source_profile.evaluate_earliest_arrival_time_at_target(0, 0), 0.5)
        found_tuples = source_profile.get_final_optimal_labels()
        self.assertEqual(len(found_tuples), 0)

    def test_target_node_not_in_walk_network(self):
        event_list_raw_data = [
            (0, 1, 0, 10, "trip_1", 1)
        ]
        transit_connections = list(map(lambda el: Connection(*el), event_list_raw_data))
        walk_speed = 2
        source_stop = 0
        target_stop = 1
        transfer_margin = 0
        start_time = 0
        end_time = 50

        walk_network = networkx.Graph()
        csa_profile = ConnectionScanProfiler(transit_connections, target_stop,
                                             start_time, end_time, transfer_margin,
                                             walk_network, walk_speed)
        csa_profile.run()
        source_profile = csa_profile.stop_profiles[source_stop]
        self.assertEqual(source_profile.evaluate_earliest_arrival_time_at_target(0, 0), 10)
        found_tuples = source_profile.get_final_optimal_labels()
        self.assertEqual(len(found_tuples), 1)


    def _assert_pareto_labels_equal(self, found_tuples, should_be_tuples):
        for found_tuple in found_tuples:
            self.assertIn(found_tuple, should_be_tuples)
        for should_be_tuple in should_be_tuples:
            self.assertIn(should_be_tuple, found_tuples)



from __future__ import print_function

import unittest

import numpy as np

from gtfspy.gtfs import GTFS
from gtfspy import shapes



class ShapesTest(unittest.TestCase):

    def test_shape_break_order(self):
        for trip_I in [
                # These trip IDs require the hsl-2015-07-12 DB.
                73775,  # Route 18 Eira -> Munkkivuori
                172258, # Route 94A in Helsinki, direction 0.
                84380,  # 36 in Helsinki.  Has lots of dead ends.
                83734,  # route 1032
                84044,  # route 1034
                240709, # 143K
                194350, # 802
                194530, # 802K
                270813, # P20
                270849, # P21
                ]:
            yield self.test_shape_break_order_1, trip_I
            pass
        #yield test_shape_break_order_1, 83734

    @unittest.skip("skipping test_shape_break_order_1")
    def test_shape_break_order_1(self, trip_I=73775):
        """This is to a bug related to shape alignment."""
        conn = GTFS('../scratch/db/hsl-2015-07-12.sqlite').conn
        cur = conn.cursor()

        cur.execute('''SELECT seq, lat, lon
                       FROM stop_times LEFT JOIN stops USING (stop_I)
                       WHERE trip_I=?
                       ORDER BY seq''',
                    (trip_I,))
        #print '%20s, %s'%(run_code, datetime.fromtimestamp(run_sch_starttime))
        stop_points = [ dict(seq=row[0],
                             lat=row[1],
                             lon=row[2])
                        for row in cur]

        # Get the shape points
        shape_id = cur.execute('''SELECT shape_id
                                  FROM trips WHERE trip_I=?''', (trip_I,)).fetchone()[0]
        shape_points = shapes.get_shape_points(cur, shape_id)
        breakpoints, badness \
              = shapes.find_segments(stop_points, shape_points)
        print(badness)
        if badness > 30:
            print("bad shape fit: %s (%s, %s)" % (badness, trip_I, shape_id))


        for b1, b2 in zip(breakpoints, sorted(breakpoints)):
            self.assertEqual(b1, b2)

    @staticmethod
    def test_interpolate_shape_times():
        shape_distances = [0, 2, 5, 10, 20, 100]
        shape_breaks = [0, 2, 5]
        stop_times = [0, 1, 20]
        result_should_be = [0, 0.4, 1, 1 + (19 * 5 / 95.), 1 + (19 * 15 / 95.), 20]

        result = shapes.interpolate_shape_times(shape_distances, shape_breaks, stop_times)
        assert len(result) == len(result_should_be)
        np.testing.assert_array_equal(result, result_should_be)

        shape_distances = [0, 1, 10]
        shape_breaks = [0, 1, 2]
        stop_times = [0, 10, 18]
        result_should_be = [0, 10, 18]
        result = shapes.interpolate_shape_times(shape_distances, shape_breaks, stop_times)
        assert len(result) == len(result_should_be)
        np.testing.assert_array_equal(result, result_should_be)



import json
import os
import shutil
import uuid

import networkx
import pandas

from gtfspy import route_types
from gtfspy.gtfs import GTFS
from gtfspy import util
from gtfspy.networks import stop_to_stop_networks_by_type, temporal_network, \
    combined_stop_to_stop_transit_network
from gtfspy.route_types import ROUTE_TYPE_TO_ZORDER


def write_walk_transfer_edges(gtfs, output_file_name):
    """
    Parameters
    ----------
    gtfs: gtfspy.GTFS
    output_file_name: str
    """
    transfers = gtfs.get_table("stop_distances")
    transfers.drop([u"min_transfer_time", u"timed_transfer"], 1, inplace=True)
    with util.create_file(output_file_name, tmpdir=True, keepext=True) as tmpfile:
        transfers.to_csv(tmpfile, encoding='utf-8', index=False)


def write_nodes(gtfs, output, fields=None):
    """
    Parameters
    ----------
    gtfs: gtfspy.GTFS
    output: str
        Path to the output file
    fields: list, optional
        which pieces of information to provide
    """
    nodes = gtfs.get_table("stops")
    if fields is not None:
        nodes = nodes[fields]
    with util.create_file(output, tmpdir=True, keepext=True) as tmpfile:
        nodes.to_csv(tmpfile, encoding='utf-8', index=False, sep=";")


def create_stops_geojson_dict(gtfs, fields=None):
    nodes = gtfs.get_table("stops")
    if fields is None:
        fields = {'name': 'stop_name', 'stop_I': 'stop_I', 'lat': 'lat', 'lon': 'lon'}
    assert (fields['lat'] == 'lat' and fields['lon'] == 'lon')
    nodes = nodes[list(fields.keys())]
    nodes.replace(list(fields.keys()), [fields[key] for key in fields.keys()], inplace=True)
    assert ('lat' in nodes.columns)
    assert ('lon' in nodes.columns)

    features = []
    for i, node_tuple in enumerate(nodes.itertuples()):
        feature = {"type": "Feature",
                   "id": str(i),
                   "geometry": {
                       "type": "Point",
                       "coordinates": [
                           node_tuple.lon,
                           node_tuple.lat
                       ]
                   },
                   "properties": {
                       "stop_I": str(node_tuple.stop_I),
                       "name": node_tuple.name
                   }
                   }
        features.append(feature)

    geojson = {
        "type": "FeatureCollection",
        "features": features
    }
    return geojson

def write_stops_geojson(gtfs, out_file, fields=None):
    """
    Parameters
    ----------
    gtfs: gtfspy.GTFS
    out_file: file-like or path to file
    fields: dict
        simultaneously map each original_name to the new_name
    Returns
    -------
    """
    geojson = create_stops_geojson_dict(gtfs, fields)
    if hasattr(out_file, "write"):
        out_file.write(json.dumps(geojson))
    else:
        with util.create_file(out_file, tmpdir=True, keepext=True) as tmpfile_path:
            tmpfile = open(tmpfile_path, 'w')
            tmpfile.write(json.dumps(geojson))


def write_combined_transit_stop_to_stop_network(gtfs, output_path, fmt=None):
    """
    Parameters
    ----------
    gtfs : gtfspy.GTFS
    output_path : str
    fmt: None, optional
        defaulting to "edg" and writing results as ".edg" files
         If "csv" csv files are produced instead    """
    if fmt is None:
        fmt = "edg"
    multi_di_graph = combined_stop_to_stop_transit_network(gtfs)
    _write_stop_to_stop_network_edges(multi_di_graph, output_path, fmt=fmt)


def write_static_networks(gtfs, output_dir, fmt=None):
    """
    Parameters
    ----------
    gtfs: gtfspy.GTFS
    output_dir: (str, unicode)
        a path where to write
    fmt: None, optional
        defaulting to "edg" and writing results as ".edg" files
         If "csv" csv files are produced instead
    """
    if fmt is None:
        fmt = "edg"
    single_layer_networks = stop_to_stop_networks_by_type(gtfs)
    util.makedirs(output_dir)
    for route_type, net in single_layer_networks.items():
        tag = route_types.ROUTE_TYPE_TO_LOWERCASE_TAG[route_type]
        file_name = os.path.join(output_dir, "network_" + tag + "." + fmt)
        if len(net.edges()) > 0:
            _write_stop_to_stop_network_edges(net, file_name, fmt=fmt)


def write_temporal_networks_by_route_type(gtfs, extract_output_dir):
    """
    Write temporal networks by route type to disk.

    Parameters
    ----------
    gtfs: gtfspy.GTFS
    extract_output_dir: str
    """
    util.makedirs(extract_output_dir)
    for route_type in route_types.TRANSIT_ROUTE_TYPES:
        pandas_data_frame = temporal_network(gtfs, start_time_ut=None, end_time_ut=None, route_type=route_type)
        tag = route_types.ROUTE_TYPE_TO_LOWERCASE_TAG[route_type]
        out_file_name = os.path.join(extract_output_dir, tag + ".tnet")
        pandas_data_frame.to_csv(out_file_name, encoding='utf-8', index=False)


def write_temporal_network(gtfs, output_filename, start_time_ut=None, end_time_ut=None):
    """
    Parameters
    ----------
    gtfs : gtfspy.GTFS
    output_filename : str
        path to the directory where to store the extracts
    start_time_ut: int | None
        start time of the extract in unixtime (seconds after epoch)
    end_time_ut: int | None
        end time of the extract in unixtime (seconds after epoch)
    """
    util.makedirs(os.path.dirname(os.path.abspath(output_filename)))
    pandas_data_frame = temporal_network(gtfs, start_time_ut=start_time_ut, end_time_ut=end_time_ut)
    pandas_data_frame.to_csv(output_filename, encoding='utf-8', index=False)


def _write_stop_to_stop_network_edges(net, file_name, data=True, fmt=None):
    """
    Write out a network

    Parameters
    ----------
    net: networkx.DiGraph
    base_name: str
        path to the filename (without extension)
    data: bool, optional
        whether or not to write out any edge data present
    fmt: str, optional
        If "csv" write out the network in csv format.
    """
    if fmt is None:
        fmt = "edg"

    if fmt == "edg":
        if data:
            networkx.write_edgelist(net, file_name, data=True)
        else:
            networkx.write_edgelist(net, file_name)
    elif fmt == "csv":
        with open(file_name, 'w') as f:
            # writing out the header
            edge_iter = net.edges_iter(data=True)
            _, _, edg_data = next(edge_iter)
            edg_data_keys = list(sorted(edg_data.keys()))
            header = ";".join(["from_stop_I", "to_stop_I"] + edg_data_keys)
            f.write(header)
            for from_node_I, to_node_I, data in net.edges_iter(data=True):
                f.write("\n")
                values = [str(from_node_I), str(to_node_I)]
                data_values = []
                for key in edg_data_keys:
                    if key == "route_I_counts":
                        route_I_counts_string = str(data[key]).replace(" ", "")[1:-1]
                        data_values.append(route_I_counts_string)
                    else:
                        data_values.append(str(data[key]))
                all_values = values + data_values
                f.write(";".join(all_values))


def create_sections_geojson_dict(G, start_time_ut=None, end_time_ut=None):
    multi_di_graph = combined_stop_to_stop_transit_network(G, start_time_ut=start_time_ut, end_time_ut=end_time_ut)
    stops = G.get_table("stops")
    stop_I_to_coords = {row.stop_I: [row.lon, row.lat] for row in stops.itertuples()}
    gjson = {"type": "FeatureCollection"}
    features = []
    gjson["features"] = features
    data = list(multi_di_graph.edges(data=True))
    data.sort(key=lambda el: ROUTE_TYPE_TO_ZORDER[el[2]['route_type']])
    for from_stop_I, to_stop_I, data in data:
        feature = {"type": "Feature"}
        geometry = {
            "type": "LineString",
            'coordinates': [stop_I_to_coords[from_stop_I], stop_I_to_coords[to_stop_I]]
        }
        feature['geometry'] = geometry
        route_I_counts = data['route_I_counts']
        route_I_counts = {str(key): int(value) for key, value in route_I_counts.items()}
        data['route_I_counts'] = route_I_counts
        properties = data
        properties['from_stop_I'] = int(from_stop_I)
        properties['to_stop_I'] = int(to_stop_I)
        feature['properties'] = data
        features.append(feature)
    return gjson

def write_sections_geojson(G, output_file, start_time_ut=None, end_time_ut=None):
    gjson = create_sections_geojson_dict(G, start_time_ut=start_time_ut, end_time_ut=end_time_ut)
    if hasattr(output_file, "write"):
        output_file.write(json.dumps(gjson))
    else:
        with open(output_file, 'w') as f:
            f.write(json.dumps(gjson))

def create_routes_geojson_dict(G):
    assert(isinstance(G, GTFS))
    gjson = {"type": "FeatureCollection"}
    features = []
    for routeShape in G.get_all_route_shapes(use_shapes=False):
        feature = {"type": "Feature"}
        geometry = {
            "type": "LineString",
            "coordinates": list(zip(routeShape['lons'], routeShape['lats']))
        }
        feature['geometry'] = geometry
        properties = {"route_type": int(routeShape['type']),
                      "route_I": int(routeShape['route_I']),
                      "route_name": str(routeShape['name'])}
        feature['properties'] = properties
        features.append(feature)
    gjson['features'] = features
    return gjson

def write_routes_geojson(G, output_file):
    gjson = create_routes_geojson_dict(G)
    if hasattr(output_file, "write"):
        output_file.write(json.dumps(gjson))
    else:
        with open(output_file, 'w') as f:
            f.write(json.dumps(gjson))
    return None



def write_gtfs(gtfs, output):
    """
    Write out the database according to the GTFS format.

    Parameters
    ----------
    gtfs: gtfspy.GTFS
    output: str
        Path where to put the GTFS files
        if output ends with ".zip" a ZIP-file is created instead.

    Returns
    -------
    None
    """
    output = os.path.abspath(output)
    uuid_str = "tmp_" + str(uuid.uuid1())
    if output[-4:] == '.zip':
        zip = True
        out_basepath = os.path.dirname(os.path.abspath(output))
        if not os.path.exists(out_basepath):
            raise IOError(out_basepath + " does not exist, cannot write gtfs as a zip")
        tmp_dir = os.path.join(out_basepath, str(uuid_str))
        # zip_file_na,e = ../out_basedir + ".zip
    else:
        zip = False
        out_basepath = output
        tmp_dir = os.path.join(out_basepath + "_" + str(uuid_str))

    os.makedirs(tmp_dir, exist_ok=True)

    gtfs_table_to_writer = {
        "agency": _write_gtfs_agencies,
        "calendar": _write_gtfs_calendar,
        "calendar_dates": _write_gtfs_calendar_dates,
        # fare attributes and fare_rules omitted (seldomly used)
        "feed_info": _write_gtfs_feed_info,
        # "frequencies": not written, as they are incorporated into trips and routes,
        # Frequencies table is expanded into other tables on initial import. -> Thus frequencies.txt is not created
        "routes": _write_gtfs_routes,
        "shapes": _write_gtfs_shapes,
        "stops": _write_gtfs_stops,
        "stop_times": _write_gtfs_stop_times,
        "transfers": _write_gtfs_transfers,
        "trips": _write_gtfs_trips,
    }

    for table, writer in gtfs_table_to_writer.items():
        fname_to_write = os.path.join(tmp_dir, table + '.txt')
        print(fname_to_write)
        writer(gtfs, open(os.path.join(tmp_dir, table + '.txt'), 'w'))

    if zip:
        shutil.make_archive(output[:-4], 'zip', tmp_dir)
        shutil.rmtree(tmp_dir)
    else:
        print("moving " + str(tmp_dir) + " to " + out_basepath)
        os.rename(tmp_dir, out_basepath)



def _remove_I_columns(df):
    """
    Remove columns ending with I from a pandas.DataFrame

    Parameters
    ----------
    df: dataFrame

    Returns
    -------
    None
    """
    all_columns = list(filter(lambda el: el[-2:] == "_I", df.columns))
    for column in all_columns:
        del df[column]


def __replace_I_with_id(gtfs, current_table, from_table_name, old_column_current, old_column_from, new_column_in_from,
                        new_column_name=None):
    if new_column_name is None:
        new_column_name = new_column_in_from
    from_table = gtfs.get_table(from_table_name)
    merged = pandas.merge(current_table, from_table, how="left",
                          left_on=old_column_current, right_on=old_column_from)
    series = pandas.Series(merged[new_column_in_from], index=current_table.index)
    current_table[new_column_name] = series


def _write_gtfs_agencies(gtfs, output_file):
    # remove agency_I
    agencies_table = gtfs.get_table("agencies")
    assert (isinstance(agencies_table, pandas.DataFrame))
    columns_to_change = {'name': 'agency_name',
                         'url': 'agency_url',
                         'timezone': 'agency_timezone',
                         'lang': 'agency_lang',
                         'phone': 'agency_phone'}
    agencies_table = agencies_table.rename(columns=columns_to_change)
    _remove_I_columns(agencies_table)
    agencies_table.to_csv(output_file, index=False)


def _write_gtfs_stops(gtfs, output_file):
    stops_table = gtfs.get_table("stops")
    assert (isinstance(stops_table, pandas.DataFrame))
    columns_to_change = {'name': 'stop_name',
                         'url': 'stop_url',
                         'lat': 'stop_lat',
                         'lon': 'stop_lon',
                         'code': 'stop_code',
                         'desc': 'stop_desc'
                         }
    stops_table = stops_table.rename(columns=columns_to_change)

    # Remove stop_I
    stop_I_to_id = {row.stop_I: row.stop_id for row in stops_table.itertuples()}
    parent_stations = []
    for stop_row in stops_table.itertuples():
        try:
            parent_station = stop_I_to_id[stop_row.parent_I]
        except KeyError:
            parent_station = ""
        parent_stations.append(parent_station)
    stops_table['parent_station'] = pandas.Series(parent_stations, index=stops_table.index)
    _remove_I_columns(stops_table)
    stops_table.to_csv(output_file, index=False)


def _write_gtfs_routes(gtfs, output_file):
    routes_table = gtfs.get_table("routes")
    columns_to_change = {'name': 'route_short_name',
                         'long_name': 'route_long_name',
                         'url': 'route_url',
                         'type': 'route_type',
                         'desc': 'route_desc',
                         'color': 'route_color',
                         'text_color': 'route_text_color'
                         }
    routes_table = routes_table.rename(columns=columns_to_change)

    # replace agency_I
    agencies_table = gtfs.get_table("agencies")
    agency_ids = pandas.merge(routes_table, agencies_table, how="left", on='agency_I')['agency_id']
    routes_table['agency_id'] = pandas.Series(agency_ids, index=routes_table.index)
    _remove_I_columns(routes_table)
    routes_table.to_csv(output_file, index=False)


def _write_gtfs_trips(gtfs, output_file):
    trips_table = gtfs.get_table("trips")
    columns_to_change = {
        'headsign': 'trip_headsign',
    }
    trips_table = trips_table.rename(columns=columns_to_change)

    __replace_I_with_id(gtfs, trips_table, 'routes', 'route_I', 'route_I', 'route_id')
    __replace_I_with_id(gtfs, trips_table, 'calendar', 'service_I', 'service_I', 'service_id')

    _remove_I_columns(trips_table)
    del [trips_table['start_time_ds']]
    del [trips_table['end_time_ds']]

    trips_table.to_csv(output_file, index=False)


def _write_gtfs_stop_times(gtfs, output_file):
    stop_times_table = gtfs.get_table('stop_times')

    columns_to_change = {
        'seq': 'stop_sequence',
        'arr_time': 'arrival_time',
        'dep_time': 'departure_time'
    }
    stop_times_table = stop_times_table.rename(columns=columns_to_change)

    # replace trip_I with trip_id
    __replace_I_with_id(gtfs, stop_times_table, 'trips', 'trip_I', 'trip_I', 'trip_id')
    __replace_I_with_id(gtfs, stop_times_table, 'stops', 'stop_I', 'stop_I', 'stop_id')

    # delete unneeded columns:
    del [stop_times_table['arr_time_hour']]
    del [stop_times_table['arr_time_ds']]
    del [stop_times_table['dep_time_ds']]
    del [stop_times_table['shape_break']]
    _remove_I_columns(stop_times_table)

    stop_times_table.to_csv(output_file, index=False)


def _write_gtfs_calendar(gtfs, output_file):
    calendar_table = gtfs.get_table('calendar')
    columns_to_change = {
        'm': 'monday',
        't': 'tuesday',
        'w': 'wednesday',
        'th': 'thursday',
        'f': 'friday',
        's': 'saturday',
        'su': 'sunday'
    }
    calendar_table = calendar_table.rename(columns=columns_to_change)
    calendar_table['start_date'] = [date.replace("-", "") for date in calendar_table['start_date']]
    calendar_table['end_date'] = [date.replace("-", "") for date in calendar_table['end_date']]
    _remove_I_columns(calendar_table)
    calendar_table.to_csv(output_file, index=False)


def _write_gtfs_calendar_dates(gtfs, output_file):
    calendar_dates_table = gtfs.get_table('calendar_dates')
    __replace_I_with_id(gtfs, calendar_dates_table, 'calendar', 'service_I', 'service_I', 'service_id')
    _remove_I_columns(calendar_dates_table)
    calendar_dates_table.to_csv(output_file, index=False)


def _write_gtfs_shapes(gtfs, ouput_file):
    shapes_table = gtfs.get_table('shapes')
    columns_to_change = {
        'lat': 'shape_pt_lat',
        'lon': 'shape_pt_lon',
        'seq': 'shape_pt_sequence',
        'd': 'shape_dist_traveled'
    }
    shapes_table = shapes_table.rename(columns=columns_to_change)
    shapes_table.to_csv(ouput_file, index=False)


def _write_gtfs_feed_info(gtfs, output_file):
    gtfs.get_table('feed_info').to_csv(output_file, index=False)


def _write_gtfs_frequencies(gtfs, output_file):
    raise NotImplementedError("Frequencies should not be outputted from GTFS as they are included in other tables.")


def _write_gtfs_transfers(gtfs, output_file):
    transfers_table = gtfs.get_table('transfers')
    __replace_I_with_id(gtfs, transfers_table, 'stops', 'from_stop_I', 'stop_I', 'stop_id', 'from_stop_id')
    __replace_I_with_id(gtfs, transfers_table, 'stops', 'to_stop_I', 'stop_I', 'stop_id', 'to_stop_id')
    _remove_I_columns(transfers_table)
    transfers_table.to_csv(output_file, index=False)


def _write_gtfs_stop_distances(gtfs, output_file):
    stop_distances = gtfs.get_table('stop_distances')
    __replace_I_with_id(gtfs, stop_distances, 'stops', 'from_stop_I', 'stop_I', 'stop_id', 'from_stop_id')
    __replace_I_with_id(gtfs, stop_distances, 'stops', 'to_stop_I', 'stop_I', 'stop_id', 'to_stop_id')
    _remove_I_columns(stop_distances)
    del stop_distances['min_transfer_time']
    del stop_distances['timed_transfer']
    stop_distances.to_csv(output_file, index=False)


# for row in stop_times_table.itertuples():
#     dep_time = gtfs.unixtime_seconds_to_gtfs_datetime(row.dep_time_ds).strftime('%H:%M%S')
#     arr_time = gtfs.unixtime_seconds_to_gtfs_datetime(row.arr_time_ds).strftime('%H:%M%S')
#     departure_times.append(dep_time)
#     arrival_times.append(arr_time)
# stop_times_table['arrival_time'] = pandas.Series(arrival_times, stop_times_table.index)
# stop_times_table['departure_time'] = pandas.Series(departure_times, stop_times_table.index)




def main():
    import argparse

    parser = argparse.ArgumentParser(description="Create network extracts from already imported GTFS files.")
    subparsers = parser.add_subparsers(dest='cmd')

    # parsing import
    parser_routingnets = subparsers.add_parser('extract_temporal', help="Direct import GTFS->sqlite")
    parser_routingnets.add_argument('gtfs', help='Input GTFS .sqlite (must end in .sqlite)')
    parser_routingnets.add_argument('basename', help='Basename for the output files')  # Parsing copy

    args = parser.parse_args()

    # if the first argument is import, import a GTFS directory to a .sqlite database.
    # Both directory and
    if args.cmd == 'extract_temporal':
        gtfs_fname = args.gtfs
        output_basename = args.basename

        from gtfspy.gtfs import GTFS
        gtfs = GTFS(gtfs_fname)

        nodes_filename = output_basename + ".nodes.csv"
        with util.create_file(nodes_filename, tmpdir=True, keepext=True) as tmpfile:
            write_nodes(gtfs, tmpfile)

        transfers_filename = output_basename + ".transfers.csv"
        with util.create_file(transfers_filename, tmpdir=True, keepext=True) as tmpfile:
            write_walk_transfer_edges(gtfs, tmpfile)

        temporal_network_filename = output_basename + ".temporal_network.csv"
        with util.create_file(temporal_network_filename, tmpdir=True, keepext=True) as tmpfile:
            write_temporal_network(gtfs, tmpfile)

    else:
        print("Unrecognized command: %s" % args.cmd)
        exit(1)


if __name__ == "__main__":
    main()





import unittest

import networkx

from gtfspy.routing.connection_scan import ConnectionScan
from gtfspy.routing.connection import Connection


class ConnectionScanTest(unittest.TestCase):

    def setUp(self):
        event_list_raw_data = [
            (1, 2, 0, 10, "trip_1", 1),
            (1, 3, 1, 10, "trip_2", 1),
            (2, 3, 10, 11, "trip_1", 2),
            (3, 4, 11, 13, "trip_1", 3),
            (3, 6, 12, 14, "trip_3", 1)
        ]
        self.transit_connections = map(lambda el: Connection(*el), event_list_raw_data)
        self.walk_network = networkx.Graph()
        self.walk_network.add_edge(4, 5, {"d_walk": 1000})
        self.walk_speed = 10
        self.source_stop = 1
        self.end_time = 20
        self.transfer_margin = 2
        self.start_time = 0 - self.transfer_margin

    def test_basics(self):
        """
        This test tests some basic features of the algorithm, such as:
        1. Transfer margins are respected.
        2. Stop labels are respected.
        3. Walk network is used properly.
        """
        csa = ConnectionScan(self.transit_connections, self.source_stop,
                             self.start_time, self.end_time,
                             self.transfer_margin, self.walk_network, self.walk_speed)
        csa.run()
        arrival_times = csa.get_arrival_times()
        self.assertEqual(arrival_times[1], self.start_time)
        self.assertEqual(arrival_times[2], 10)
        self.assertEqual(arrival_times[3], 10)
        self.assertEqual(arrival_times[4], 13)
        self.assertEqual(arrival_times[5], 13 + 100)
        self.assertEqual(arrival_times[6], 14)
        self.assertEqual(arrival_times[7], float('inf'))
        self.assertGreater(csa.get_run_time(), 0)

    def test_change_starttime(self):
        start_time = 1 - self.transfer_margin
        csa = ConnectionScan(self.transit_connections, self.source_stop,
                             start_time, self.end_time,
                             self.transfer_margin, self.walk_network, self.walk_speed)
        csa.run()
        arrival_times = csa.get_arrival_times()
        self.assertEqual(arrival_times[1], start_time)
        self.assertEqual(arrival_times[2], float('inf'))
        self.assertEqual(arrival_times[3], 10)
        self.assertEqual(arrival_times[4], float('inf'))
        self.assertEqual(arrival_times[5], float('inf'))
        self.assertEqual(arrival_times[6], 14)
        self.assertEqual(arrival_times[7], float('inf'))

    def test_change_endtime(self):
        end_time = 11
        csa = ConnectionScan(self.transit_connections, self.source_stop,
                             self.start_time, end_time,
                             self.transfer_margin, self.walk_network, self.walk_speed)
        csa.run()
        arrival_times = csa.get_arrival_times()
        self.assertEqual(arrival_times[1], self.start_time)
        self.assertEqual(arrival_times[2], 10)
        self.assertEqual(arrival_times[3], 10)
        self.assertEqual(arrival_times[4], 13)
        self.assertEqual(arrival_times[5], 13 + 100)
        self.assertEqual(arrival_times[6], float('inf'))
        self.assertEqual(arrival_times[7], float('inf'))

    def test_starts_with_walk(self):
        end_time = 11
        event_list_raw_data = [
            (1, 2, 0, 10, "trip_1", 1)
        ]
        transit_connections = map(lambda el: Connection(*el), event_list_raw_data)
        walk_network = networkx.Graph()
        walk_network.add_edge(1, 2, {"d_walk": 10})
        walk_speed = 10
        source_stop = 1
        start_time = 0
        transfer_margin = 0
        csa = ConnectionScan(transit_connections, source_stop,
                             start_time, end_time,
                             transfer_margin, walk_network, walk_speed)
        csa.run()
        arrival_times = csa.get_arrival_times()
        self.assertEqual(arrival_times[1], start_time)
        self.assertEqual(arrival_times[2], 1)









